"""
DiffProcessor Plugin
####################

Processor plugin to run diff against previous task's results saved
using ``ToFileProcessor``.

DiffProcessor Sample Usage
==========================

Code to demonstrate how to use ``DiffProcessor`` plugin::

    from nornir import InitNornir
    from nornir_netmiko import netmiko_send_command
    from nornir_salt import DiffProcessor

    nr = InitNornir(config_file="config.yaml")

    nr_with_processor = nr.with_processors([
        DiffProcessor(diff="config")
    ])

    nr_with_processor.run(
        task=netmiko_send_command,
        command_string="show run"
    )

DiffProcessor reference
=========================

.. autofunction:: nornir_salt.plugins.processors.DiffProcessor.DiffProcessor
"""
import logging
import os
import json
import traceback
import difflib
import re

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task


log = logging.getLogger(__name__)


class DiffProcessor:
    """
    DiffProcessor can report difference between current task's results and previously
    saved results.

    :param diff: (str) filegroup name to run diff for
    :param base_url: (str) OS path to folder with files, default "/var/nornir-salt/"
    :param last: (int) default is 1, file index to compare with
    :param ignore_lines: (list) list of regular expression pattern to filter lines through,
        if line matches any of the patterns, it is ignored by removing it; by default any
        line that contains only space characters ignored
    :param remove_patterns: (list) list of regular expression pattern to remove from lines
    :param in_diff: (bool) default is False, if True uses ``Result`` object ``diff`` attribute
        to store diff results, otherwise uses task's ``Result.result`` attribute to store diffs
    :param index: (str) ``ToFileProcessor`` index file name to read files info from

    ``ignore_lines`` and ``remove_patterns`` arguments exists to clean difference results,
    for instance by ignoring timestamps, counters or other uninteresting data.
    """

    def __init__(
        self,
        diff,
        base_url="/var/nornir-salt/",
        last=1,
        in_diff=False,
        ignore_lines=None,
        remove_patterns=None,
        index=None,
    ):
        self.diff = diff
        self.base_url = base_url
        self.last = last
        self.ignore_lines = ignore_lines or [r"^\s*[\n\r]+$"]
        self.remove_patterns = remove_patterns or []
        self.in_diff = in_diff
        self.index = index or "common"

        self.aliases_file = os.path.join(
            base_url, "tf_index_{}.json".format(self.index)
        )
        self.aliases_data = (
            {}
        )  # dictionary of {diff name: {hostname: [{filename: str, tasks: {task_name: file_span}}]}}

        self._load_aliases()

    def _load_aliases(self):
        # load aliases data from aliases file
        if os.path.exists(self.aliases_file):
            with open(self.aliases_file, mode="r", encoding="utf-8") as f:
                self.aliases_data = json.loads(f.read())

    def _filter_ignore_lines(self, data):
        """
        Helper function to filter lines using regex from ignore_lines list

        :param data: (list) list of lines to filter
        """
        return [
            ln
            for ln in data
            if not any(
                map(lambda pt: True if re.search(pt, ln) else False, self.ignore_lines)
            )
        ]

    def _remove_patterns(self, data):
        """
        Helper function to remove patterns using regex from remove_patterns list

        :param data: (list) list of lines to remove patterns from
        """
        for index, ln in enumerate(data):
            for pattern in self.remove_patterns:
                ln = re.sub(pattern, "", ln)
            data[index] = ln
        return data

    def _run_diff(self, prev_result, new_result, fromfile, tofile):
        """
        Helper function to run diff

        :param prev_result: (str) multiline string to run diff for
        :param new_result: (str) multiline string to run diff for
        :param fromfile: (str) from file name to use with difflib
        :param tofile: (str) to file name to use with difflib
        """
        # filter data through ignore_lines patterns
        if self.ignore_lines:
            prev_result = self._filter_ignore_lines(
                prev_result.splitlines(keepends=True)
            )
            new_result = self._filter_ignore_lines(new_result.splitlines(keepends=True))

        # clean up data using remove_patterns
        if self.remove_patterns:
            prev_result = self._remove_patterns(prev_result)
            new_result = self._remove_patterns(new_result)

        return difflib.unified_diff(
            prev_result, new_result, fromfile=fromfile, tofile=tofile
        )

    def task_started(self, task: Task) -> None:
        pass  # ignore

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """Diff files with current task result"""
        # check if has failed tasks, do nothing in such a case
        if result.failed:
            log.error("nornir_salt:DiffProcessor do nothing, return, has failed tasks")
            return

        try:
            # get previous results data
            index = min(self.last - 1, len(self.aliases_data[self.diff][host.name]) - 1)
            prev_res_alias_data = self.aliases_data[self.diff][host.name][index]
            prev_res_filename = prev_res_alias_data["filename"]

            # open previous results file
            with open(prev_res_filename, mode="r", encoding="utf-8") as f:
                prev_result = f.read()

            # run diff for each task
            for i in result:
                # check if need to skip this task results
                if hasattr(i, "skip_results") and i.skip_results is True:
                    continue

                # check if task results exists
                if i.name not in prev_res_alias_data["tasks"]:
                    i.diff = "'{}' task results not in '{}''".format(
                        i.name, prev_res_filename
                    )
                    continue

                # form new results
                if isinstance(i.result, (str, int, float, bool)):
                    new_result = str(i.result) + "\n"
                # convert structured data to json
                else:
                    new_result = (
                        json.dumps(
                            i.result, sort_keys=True, indent=4, separators=(",", ": ")
                        )
                        + "\n"
                    )

                # run diff using portion of prev_result file with given task results only
                spans = prev_res_alias_data["tasks"][i.name]["span"]
                difference = self._run_diff(
                    prev_result=prev_result[spans[0] : spans[1]],
                    new_result=new_result,
                    fromfile=prev_res_filename,
                    tofile="current",
                )
                diff_res = "".join(difference)
                diff_res = diff_res if diff_res else True
                if self.in_diff:
                    i.diff = diff_res
                else:
                    i.result = diff_res
        except:
            log.error(
                "nornir-salt:DiffProcessor host {} error:\n{}".format(
                    host.name, traceback.format_exc()
                )
            )

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore subtasks

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass  # ignore subtasks

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass



"""
ncclient_call
#############

`Ncclient <https://github.com/ncclient>`_ is a popular library to interact with
devices using NETCONF, this plugin is a wrapper around ncclient connection
manager object.

NETCONF protocol has a specific set of RPC calls available for use, rather
than coding separate task for each of them, ``ncclient_call`` made to execute
any arbitrary method supported by manager object plus a set of additional
helper methods to extend Ncclient library functionality.

ncclient_call sample usage
==========================

Sample code to run ``ncclient_call`` task::

    from nornir import InitNornir
    from nornir_salt import ncclient_call

    nr = InitNornir(config_file="config.yaml")

    output = nr.run(
        task=ncclient_call,
        call="get_config",
        source="running"
    )

ncclient_call returns
=====================

Returns XML text string by default, but can return XML data transformed
in JSON, YAML or Python format.

ncclient_call reference
=======================

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call.ncclient_call

additional methods reference
============================

ncclient_call - dir
-------------------
.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_dir

ncclient_call - help
--------------------
.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_help

ncclient_call - server_capabilities
-----------------------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_server_capabilities

ncclient_call - connected
-------------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_connected

ncclient_call - transaction
---------------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_transaction
"""
import traceback
import logging
import time

from nornir.core.task import Result, Task
from nornir_salt.plugins.connections.NcclientPlugin import CONNECTION_NAME

log = logging.getLogger(__name__)

try:
    import lxml.etree as etree  # nosec

    HAS_LXML = True
except ImportError:
    HAS_LXML = False

try:
    from ncclient.manager import OPERATIONS

    HAS_NCCLIENT = True
except ImportError:
    HAS_NCCLIENT = False

try:
    # this import should work for ncclient >=0.6.10
    from ncclient.operations import GenericRPC

except ImportError:
    # for ncclient<0.6.10 need to reconstruct GenericRPC class
    if HAS_NCCLIENT:

        from ncclient.operations import RPC

        class GenericRPC(RPC):
            def request(self, data, *args, **kwargs):
                """
                :param data: (str) rpc xml string

                Testing:

                * Arista cEOS - not working, transport session closed error
                * Cisco IOS XR - working
                """
                ele = etree.fromstring(data.encode("UTF-8"))  # nosec
                return self._request(ele)


def _call_transaction(manager, *args, **kwargs):
    """
    Function to edit device configuration in a reliable fashion using
    capabilities advertised by NETCONF server.

    :param target: (str) name of datastore to edit configuration for, if no
        ``target`` argument provided and device supports candidate datastore uses
        ``candidate`` datastore, uses ``running`` datastore otherwise
    :param config: (str) configuration to apply
    :param format: (str) configuration string format, default is xml
    :param confirmed: (bool) if True (default) uses commit confirmed
    :param commit_final_delay: (int) time to wait before doing final commit after
        commit confirmed, default is 1 second
    :param confirm_delay: (int) device commit confirmed rollback delay, default 60 seconds
    :param validate: (bool) if True (default) validates candidate configuration before commit
    :returns result: (list) list of steps performed with details

    Function work flow:

    1. Lock target configuration datastore
    2. If server supports it - Discard previous changes if any
    3. Edit configuration
    4. If server supports it - validate configuration if ``validate`` argument is True
    5. If server supports it - do commit confirmed if ``confirmed`` argument is True
    6. If server supports it - do commit operation
    7. Unlock target configuration datastore
    8. If server supports it - discard all changes if any of steps 3, 4, 5 or 6 fail
    9. Return results list of dictionaries keyed by step name
    """
    result = []
    failed = False
    commit_final_delay = int(kwargs.get("commit_final_delay", 1))
    confirm_delay = str(
        kwargs.get("confirm_delay", 60)
    )  # ncclient expects timeout to be a string

    # get capabilities
    can_validate = ":validate" in manager.server_capabilities
    can_commit_confirmed = ":confirmed-commit" in manager.server_capabilities
    has_candidate_datastore = ":candidate" in manager.server_capabilities

    # decide on target configuration datastore
    kwargs["target"] = kwargs.get(
        "target", "candidate" if has_candidate_datastore else "running"
    )

    # execute transaction
    with manager.locked(target=kwargs["target"]):
        if has_candidate_datastore and kwargs["target"] == "candidate":
            r = manager.discard_changes()
            result.append(
                {"discard_changes": etree.tostring(r._root, pretty_print=True)}
            )
        try:
            r = manager.edit_config(
                config=kwargs["config"],
                target=kwargs["target"],
                format=kwargs.get("format", "xml"),
            )
            result.append({"edit_config": etree.tostring(r._root, pretty_print=True)})
            # validate configuration
            if can_validate and kwargs.get("validate", True):
                r = manager.validate(source=kwargs["target"])
                result.append({"validate": etree.tostring(r._root, pretty_print=True)})
            if kwargs["target"] == "candidate" and has_candidate_datastore:
                # run commit confirmed
                if can_commit_confirmed and kwargs.get("confirmed", True):
                    pid = "dob04041989"
                    r = manager.commit(
                        confirmed=True, timeout=confirm_delay, persist=pid
                    )
                    result.append(
                        {"commit_confirmed": etree.tostring(r._root, pretty_print=True)}
                    )
                    # run final commit
                    time.sleep(commit_final_delay)
                    r = manager.commit(confirmed=True, persist_id=pid)
                    result.append(
                        {"commit": etree.tostring(r._root, pretty_print=True)}
                    )
                # run normal commit
                else:
                    r = manager.commit()
                    result.append(
                        {"commit": etree.tostring(r._root, pretty_print=True)}
                    )
        except:
            tb = traceback.format_exc()
            log.error("nornir_salt:ncclient_call transaction error: {}".format(tb))
            result.append({"error": tb})
            if has_candidate_datastore and kwargs["target"] == "candidate":
                r = manager.discard_changes()
                result.append(
                    {"discard_changes": etree.tostring(r._root, pretty_print=True)}
                )
            failed = True

    return result, failed


def _call_server_capabilities(manager, *args, **kwargs):
    """Helper function to get server capabilities"""
    return [c for c in manager.server_capabilities], False


def _call_connected(manager, *args, **kwargs):
    """Helper function to get connected status"""
    return manager.connected, False


def _call_dir(manager, *args, **kwargs):
    """Function to return alist of available methods/operations"""
    methods = (
        list(dir(manager))
        + list(manager._vendor_operations.keys())
        + list(OPERATIONS.keys())
        + ["dir", "help", "transaction"]
    )
    result = sorted(
        [m for m in set(methods) if (not m.startswith("_") and not m.isupper())]
    )
    return result, False


def _call_help(manager, method_name, *args, **kwargs):
    """
    Helper function to return docstring for requested method

    :param method_name: (str) name of method or function to return docstring for
    """
    if "_call_{}".format(method_name) in globals():
        function_obj = globals()["_call_{}".format(method_name)]
    else:
        function_obj = getattr(manager, method_name)
    h = function_obj.__doc__ if hasattr(function_obj, "__doc__") else ""
    return h, False


def ncclient_call(task: Task, call: str, *args, **kwargs) -> Result:
    """
    Task to handle a call of NCClient manager object methods

    :param call: (str) ncclient manager object method to call
    :param arg: (list) any ``*args`` to use with call method
    :param kwargs: (dict) any ``**kwargs`` to use with call method
    """
    # run sanity check
    if not HAS_NCCLIENT:
        return Result(
            host=task.host, failed=True, exception="No Ncclient found, is it installed?"
        )

    # initiate parameters
    failed = False
    task.name = call

    # get rendered data if any
    if "__task__" in task.host.data:
        kwargs.update(task.host.data["__task__"])

    # check if filter formed properly - as per
    # https://ncclient.readthedocs.io/en/latest/manager.html#filter-params
    # filter should be a tuple of (type, criteria)
    if kwargs.get("filter"):
        if isinstance(kwargs["filter"], list):
            kwargs["filter"] = tuple(kwargs["filter"])
        elif isinstance(kwargs["filter"], str):
            kwargs["filter"] = tuple([kwargs.pop("ftype", "subtree"), kwargs["filter"]])

    # get Ncclient NETCONF connection object
    manager = task.host.get_connection(CONNECTION_NAME, task.nornir.config)

    # add generic RPC operation to Ncclient manager object to support RPC call
    manager._vendor_operations.setdefault("rpc", GenericRPC)

    log.debug(
        "nornir_salt:ncclient_call calling '{}' with args: '{}'; kwargs: '{}'".format(
            call, args, kwargs
        )
    )

    # check if need to call one of helper function
    if "_call_{}".format(call) in globals():
        result, failed = globals()["_call_{}".format(call)](manager, *args, **kwargs)
    # call manager object method otherwise
    else:
        result = getattr(manager, call)(*args, **kwargs)

    # format results
    if hasattr(result, "_root"):
        result = etree.tostring(result._root, pretty_print=True).decode()
    elif isinstance(result, (list, dict, bool)):
        pass
    else:
        result = str(result)

    return Result(host=task.host, result=result, failed=failed)



"""
NcclientPlugin
##############

Ncclient library connection plugin to interact with devices over NETCONF.

NcclientPlugin reference
========================

.. autofunction:: nornir_salt.plugins.connections.NcclientPlugin.NcclientPlugin
"""
from typing import Any, Dict, Optional
from pathlib import Path
from nornir.core.configuration import Config

try:
    from ncclient import manager

    HAS_NCCLIENT = True
except ImportError:
    HAS_NCCLIENT = False

CONNECTION_NAME = "ncclient"


class NcclientPlugin:
    """
    Full list of inventory extras see Ncclient Docs for ``manager.connect``
    method.

    Example on how to configure a device to use netconfig without using
    an ssh agent and without verifying the keys::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          port: 2022
          connection_options:
            ncclient:
              extras:
                allow_agent: False
                hostkey_verify: False

    Anything under inventory ``extras`` section passed on to Ncclient
    ``manager.connect`` call.
    """  # noqa

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        extras = extras or {}

        parameters: Dict[str, Any] = {
            "host": hostname,
            "username": username,
            "password": password,
            "port": port or 830,
        }

        if "ssh_config" not in extras:
            ssh_config_file = Path(configuration.ssh.config_file)
            if ssh_config_file.exists():
                parameters["ssh_config"] = ssh_config_file

        parameters.update(extras)

        self.connection = manager.connect(**parameters)

    def close(self) -> None:
        self.connection.close_session()



"""
pyats_send_config
#######################

This task plugin relies on Genie device conection ``config`` method
to send configuration commands to devices over SSH or Telnet.

This task plugin applies device configuration following this sequence:

- Retrieve and use, if any, per-host configuration rendered by SaltStack from host's
  inventory data ``task.host.data["__task__"]["commands"]`` or
  ``task.host.data["__task__"]["filename"]`` locations, use configuration provided
  by ``config`` argument otherwise
- If configuration is a multi-line string, split it to a list of commands
- Check if device in enable mode, if not enter device enabled mode if device supports it
- Push configuration commands to device using ``send_config_set`` Netmiko connection's method,
  if ``batch`` argument given, pushes commands in batches
- If ``commit`` argument provided, perform configuration commit if device supports it
- If ``commit_final_delay`` argument provided, wait for a given timer and perform final commit
- Exit device configuration mode and return configuration results

Dependencies:

* `PyATS library <https://pypi.org/project/pyats/>`_ required
* `Genie library <https://pypi.org/project/genie/>`_ required

Sample Usage
============

Code to invoke ``pyats_send_config`` task::

    from nornir_salt import pyats_send_config

    output = nr.run(
        task=pyats_send_config,
        commands=["sinterface loopback 0", "description 'configured by script'"]
    )

``pyats_send_config`` returns Nornir results object with task name set
to ``pyats_send_config`` and results containing configuration commands
applied to device.

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.pyats_send_config.pyats_send_config
"""
import logging
import traceback
from nornir.core.task import Result, Task

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "pyats"


def pyats_send_config(task: Task, config: str = None, **kwargs):
    """
    Salt-nornir Task function to send configuration to devices using
    ``nornir_netmiko.tasks.pyats_send_config`` plugin.

    Device ``configure`` method does not support specifying connection to use to
    send configuration via.

    :param config: (str or list) configuration string or list of commands to send to device
    :param kwargs: (dict) any additional ``**kwargs`` for device connection ``configure`` method
    :return result: Nornir result object with task execution results

    Device ``configure`` method supports below additional arguments that can be passed
    via ``**kwargs``:

    :param reply: Addition Dialogs for interactive config commands.
    :param timeout: Timeout value in sec, Default Value is 30 sec
    :param error_pattern: list of regex to detect command errors
    :param target: Target RP where to execute service, for DualRp only
    :param lock_retries: retry times if config mode is locked, default is 0
    :param lock_retry_sleep: sleep between retries, default is 2 sec
    :param bulk: If False, send all commands in one sendline,
        If True, send commands in chunked mode, default is False
    :param bulk_chunk_lines: maximum number of commands to send per chunk,
        default is 50, 0 means to send all commands in a single chunk
    :param bulk_chunk_sleep: sleep between sending command chunks, default is 0.5 sec
    """
    # run sanity check
    if kwargs.get("dry_run"):
        raise ValueError("pyats_send_config does not support dry_run")

    task.name = "pyats_send_config"
    task_result = Result(host=task.host, result=[], changed=True)

    # get PyATS testbed, device object
    testbed = task.host.get_connection(CONNECTION_NAME, task.nornir.config)
    device = testbed.devices[task.host.name]

    # get configuration from host data if any
    if "config" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["config"]
    elif "commands" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["filename"]

    # transform configuration to a list if string given
    if isinstance(config, str):
        config = config.splitlines()

    # send config
    try:
        task_result.result = device.configure(config, **kwargs)
    except:
        log.exception("nornir-salt:pyats_send_config configure error")
        task_result.failed = True
        task_result.exception = traceback.format_exc()
        task_result.changed = False

    return task_result



"""
scrapli_send_commands
#####################

This plugin uses ``nornir-scrapli`` ``send_command`` task to send multiple commands
to devices pre-processing commands accordingly.

Pre-processing includes:

- Check and if any - retrieve per-host commands from host's inventory data
  ``task.host.data["__task__"]["commands"]`` or from ``task.host.data["__task__"]["filename"]``
- If command is a multi-line string, split it to individual lines or form a list with single command
- Iterate over commands list and remove empty strings
- Iterate over commands and replace ``new_line_char`` with ``\\n`` new line

Next, commands send one by one to device using ``send_command`` task plugin,
each task named after the command being sent. Sleep for given ``interval`` between
sending commands.

Dependencies:

* `nornir-scrapli module <https://pypi.org/project/nornir-scrapli/>`_ required

Sample Usage
============

Code to invoke ``scrapli_send_commands`` task::

    from nornir import InitNornir
    from nornir_salt import scrapli_send_commands

    nr = InitNornir(config_file="config.yaml")

    output = nr.run(
        task=scrapli_send_commands,
        commands=["show run", "show clock"]
    )

Task scrapli_send_commands returns Nornir results object with individual tasks
names set equal to commands sent to device.

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.scrapli_send_commands.scrapli_send_commands
"""
import time
from nornir.core.task import Result, Task

try:
    from nornir_scrapli.tasks import send_command

    HAS_SCRAPLI = True
except ImportError:
    HAS_SCRAPLI = False

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "scrapli"


def scrapli_send_commands(
    task: Task, commands=None, interval=0.01, new_line_char="_br_", **kwargs
):
    """
    Nornir Task function to send show commands to devices using
    ``nornir_scrapli.tasks.send_command`` plugin

    Per-host ``commands`` can be provided using host's object ``data`` attribute
    with ``__task__`` key with value set to dictionary with ``commands`` key
    containing a list of or a multiline string of commands to send to device, e.g.::

        print(host.data["__task__"]["commands"])

        ["ping 1.1.1.1 source 1.1.1.2", "show clock"]

    :param commands: (list or str) list or multiline string of commands to send to device
    :param interval: (int) interval between sending commands, default 0.01s
    :param new_line_char: (str) characters to replace in commands with new line ``\\n``
        before sending command to device, default is ``_br_``, useful to simulate enter key
    :param kwargs: (dict) any additional arguments to pass to scrapli send_command
    :return result: Nornir result object with task execution results
    """
    # run sanity check
    if not HAS_SCRAPLI:
        return Result(
            host=task.host,
            failed=True,
            exception="No nornir_scrapli found, is it installed?",
        )

    commands = commands or []

    # run interval sanity check
    interval = interval if isinstance(interval, (int, float)) else 0.01

    # get per-host commands if any
    if "commands" in task.host.data.get("__task__", {}):
        if commands:
            for c in task.host.data["__task__"]["commands"]:
                if c not in commands:
                    commands.append(c)
        else:
            commands = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        commands = task.host.data["__task__"]["filename"]

    # normilize commands to a list
    if isinstance(commands, str):
        commands = commands.splitlines()

    # remove empty lines/commands that can left after rendering
    commands = [c for c in commands if c.strip()]

    # iterate over commands and see if need to add empty line - hit enter
    commands = [
        c.replace(new_line_char, "\n") if new_line_char in c else c for c in commands
    ]

    # send commands to device
    for command in commands:
        task.run(task=send_command, command=command, name=command.strip(), **kwargs)
        time.sleep(interval)

    # set skip_results to True, for ResultSerializer to ignore
    # results for grouped task itself, which are usually None
    return Result(host=task.host, skip_results=True)



from .NcclientPlugin import NcclientPlugin
from .HTTPPlugin import HTTPPlugin
from .PyGNMIPlugin import PyGNMIPlugin
from .ConnectionsPool import ConnectionsPool
from .PyATSUnicon import PyATSUnicon

__all__ = (
    "NcclientPlugin",
    "HTTPPlugin",
    "PyGNMIPlugin",
    "ConnectionsPool",
    "PyATSUnicon",
)



"""
InventoryFun
############

Function to interact with in memory Nornir Inventory data.

InventoryFun Sample Usage
===========================

Sample code to invoke ``InventoryFun`` function to ``create`` new host::

    from nornir_salt import InventoryFun

    host_data = {
        "name": "IOL3",
        "hostname": "192.168.217.99",
        "platform": "ios",
        "groups": ["lab", "bma"],
        "connection_options": {
            "napalm": {
                "port": 2022,
                "extras": {
                    "foo": "bar"
                }
            }
        }
    }

    res = InventoryFun(nr, call="create_host", **host_data)
    # or res = InventoryFun(nr, "create_host", **host_data)

Sample code to invoke ``InventoryFun`` function to ``update`` existing host::

    from nornir_salt import InventoryFun

    host_data = {
        "name": "IOL2",
        "hostname": "192.168.217.99",
        "platform": "ios_xe",
        "username": "new_username",
        "password": "new_password",
        "port": 123,
        "groups": ["lab", "bma"],
        "data": {
            "made_by": "humans",
        },
        "connection_options": {
            "napalm": {
                "port": 2022,
                "extras": {
                    "foo": "bar1"
                }
            }
        }
    }

    res = InventoryFun(nr, call="update_host", groups_action="append", **host_data)
    # or res = InventoryFun(nr, "update_host", groups_action="append", **host_data)

Sample code to invoke ``InventoryFun`` function to ``delete`` existing host::

    from nornir_salt import InventoryFun

    res = InventoryFun(nr, call="delete_host", name="IOL2")
    # or res = InventoryFun(nr, "delete_host", name="IOL2")

Sample code to invoke ``InventoryFun`` function to bulk ``load`` from list::

    from nornir_salt import InventoryFun

    data = [
        {
            "call": "create_host",
            "name": "IOL3",
            "hostname": "192.168.217.99",
            "platform": "ios",
            "groups": ["lab", "bma"],
        },
        {
            "call": "delete_host",
            "name": "IOL2",
        },
        {
            "call": "update_host",
            "name": "IOL1",
            "hostname": "1.2.3.4",
            "platform": "ios_xe",
            "groups": ["bma"],
            "groups_action": "remove"
        },
        {
            "call": "create",
            "name": "IOL4",
            "hostname": "192.168.217.4",
            "platform": "iosxr",
            "groups": ["lab"],
        },
    ]

    res = InventoryFun(nr, call="load", data=data)
    # or res = InventoryFun(nr, "load", data=data)

InventoryFun Reference
======================

.. autofunction:: nornir_salt.plugins.functions.InventoryFun.InventoryFun

Call Functions Reference
========================

.. autofunction:: nornir_salt.plugins.functions.InventoryFun._create_host
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._read_host
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._read_inventory
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._update_host
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._delete_host
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._load
.. autofunction:: nornir_salt.plugins.functions.InventoryFun._list_hosts
"""
import logging

from .FFun import FFun
from nornir.core.inventory import Host, ConnectionOptions

log = logging.getLogger(__name__)


def _create_host(nr, name, groups=None, connection_options=None, **kwargs):
    """
    Function to add new host in inventory or replace existing host.

    :param nr: (obj) Nornir object
    :param name: (str) host name
    :param groups: (list) list of host's parent group names
    :param connection_options: (dict) connection options dictionary
    :param kwargs: (dict) host base attributes such as hostname, port,
        username, password, platform, data
    :return: True on success

    If group given in ``groups`` list does not exist, no error raised, it simply skipped.
    """
    groups = groups or []
    connection_options = connection_options or {}

    # add new host or replace existing host completely
    nr.inventory.hosts[name] = Host(
        name=name,
        hostname=kwargs.pop("hostname", name),
        defaults=nr.inventory.defaults,
        connection_options={
            cn: ConnectionOptions(
                hostname=c.get("hostname"),
                port=c.get("port"),
                username=c.get("username"),
                password=c.get("password"),
                platform=c.get("platform"),
                extras=c.get("extras"),
            )
            for cn, c in connection_options.items()
        },
        groups=[
            nr.inventory.groups[group_name]
            for group_name in groups
            if group_name in nr.inventory.groups
        ],
        **kwargs
    )

    return True


def _read_host(nr, **kwargs):
    """
    Function to return inventory content for host(s) using FFun
    function to filter hosts.

    :param nr: (obj) Nornir object
    :param kwargs: (dict) arguments to use with FFun function to filter hosts
    :return: (dict) host(s) inventory dictionary keyed by host name(s)
    """
    hosts = FFun(nr, kwargs=kwargs)
    return hosts.inventory.dict()["hosts"]


def _read_inventory(nr, **kwargs):
    """
    Function to return inventory content using FFun function to filter hosts.

    :param nr: (obj) Nornir object
    :param kwargs: (dict) arguments to use with FFun function to filter hosts
    :return: (dict) inventory dictionary with groups, defaults and hosts
    """
    hosts = FFun(nr, kwargs=kwargs)
    return hosts.inventory.dict()


def _update_host(
    nr,
    name,
    groups=None,
    connection_options=None,
    data=None,
    groups_action="append",
    **kwargs
):
    """
    Function to update host's inventory.

    :param nr: (obj) Nornir object
    :param name: (str) host name
    :param groups: (list) list of host's parent group names
    :param groups_action: (str) what to do with groups - ``append`` (default),
        ``insert`` (prepend) or ``remove``
    :param connection_options: (dict) connection options dictionary
    :param data: (dict) dictionary with host's data
    :param kwargs: (dict) host base attributes such as hostname, port,
        username, password, platform
    :return: True on success

    ``data`` and ``connection_options`` replace existing values for top keys
    similar to dictionary ``update`` method, no recursive merge performed.

    If group given in ``groups`` list does not exist, no error raised, it simply skipped.
    """
    groups = groups or []
    connection_options = connection_options or {}
    data = data or {}

    # do sanity checks
    if groups_action not in ["append", "insert", "remove"]:
        raise ValueError(
            "nornir-salt:InventoryFun groups_action should be append, insert or remove, not '{}'".format(
                groups_action
            )
        )

    # get host object to update
    try:
        host_obj = nr.inventory.hosts[name]
    except KeyError:
        return False
    existing_groups = [g.name for g in host_obj.groups]

    # update base attributes
    host_obj.hostname = kwargs.get("hostname", host_obj.hostname)
    host_obj.platform = kwargs.get("platform", host_obj.platform)
    host_obj.port = kwargs.get("port", host_obj.port)
    host_obj.username = kwargs.get("username", host_obj.username)
    host_obj.password = kwargs.get("password", host_obj.password)

    # update data
    host_obj.data.update(data)

    # update connection options
    for cn, c in connection_options.items():
        host_obj.connection_options[cn] = ConnectionOptions(
            hostname=c.get("hostname"),
            port=c.get("port"),
            username=c.get("username"),
            password=c.get("password"),
            platform=c.get("platform"),
            extras=c.get("extras"),
        )

    # update groups
    for group_name in groups:
        if group_name not in nr.inventory.groups:
            continue
        if groups_action == "append" and group_name not in existing_groups:
            host_obj.groups.append(nr.inventory.groups[group_name])
        elif groups_action == "insert" and group_name not in existing_groups:
            host_obj.groups.insert(0, nr.inventory.groups[group_name])
        elif groups_action == "remove" and group_name in existing_groups:
            host_obj.groups.remove(nr.inventory.groups[group_name])

    return True


def _delete_host(nr, name):
    """
    Function to delete host from inventory.

    :param nr: (obj) Nornir object
    :param name: (str or list) host name or a list of host names to delete
    :return: True on success
    """
    names = [name] if isinstance(name, str) else name
    for n in names:
        _ = nr.inventory.hosts.pop(n, None)
    return True


def _load(nr, data):
    """
    Accept a list of items, where each item executed sequentially
    to perform one of the operations to create, update or delete.

    :param data: (list) list of dictionary items to work with
    :param nr: (obj) Nornir Object

    Each list dictionary item should contain ``call`` key holding the name
    of function to call, rest of the dictionary used as a ``**kwargs`` with
    specidfied call function.
    """
    for item in data:
        fun_name = item.pop("call")
        fun_dispatcher[fun_name](nr, **item)
    return True


def _list_hosts(nr, **kwargs):
    """
    Function to return a list of host names contained in inventory.

    Supports filtering using FFun function.

    :param nr: (obj) Nornir object
    :param kwargs: (dict) FFun function arguments to filter hosts
    :return: (list) list of host names
    """
    hosts = FFun(nr, kwargs=kwargs)
    return list(hosts.inventory.hosts.keys())


fun_dispatcher = {
    "create_host": _create_host,
    "update_host": _update_host,
    "delete_host": _delete_host,
    "read_host": _read_host,
    "create": _create_host,
    "update": _update_host,
    "delete": _delete_host,
    "read": _read_host,
    "read_inventory": _read_inventory,
    "load": _load,
    "list_hosts": _list_hosts,
}


def InventoryFun(nr, call, **kwargs):
    """
    Dispatcher function to execute one of ``call`` functions.

    :param nr: (obj) Nornir object
    :param call: (str) name of function to call
    :param kwargs: (dict) arguments to pass on to call function
    :returns: call function results

    Supported ``call`` function values:

    - ``create_host`` or ``create`` - calls ``_create_host``, creates new host or replaces existing host object
    - ``read_host`` or ``read`` - calls ``_read_host``, read host inventory content
    - ``update_host`` or ``update`` - calls ``_update_host``, non recursively update host attributes
    - ``delete_host`` or ``delete`` - calls ``_delete_host``, deletes host objcet from Nornir Inventory
    - ``load`` - calls ``_load``, to simplify calling multiple functions
    - ``read_inventory`` - calls ``_read_inventory``, read inventory content for groups, default and hosts
    - ``list_hosts`` - calls ``_list_hosts``, return a list of inventory's host names
    """
    return fun_dispatcher[call](nr, **kwargs)



"""
RetryRunner plugin
##################

RetryRunner plugin implements retry logic to improve task execution reliability.

.. warning:: For grouped tasks need to explicitly provide `connection_name` attribute
    such as `netmiko`, `napalm`, `scrapli`. Specifying `connection_name` attribute for
    standalone tasks not required. Lack of `connection_name` attribute will result in skipping
    connection retry logic and connections to all hosts initiated simultaneously up to the
    number of `num_workers`

Primary motivation for RetryRunner is to make Nornir task execution as reliable as possible
through retry mechanisms.

RetryRunner Architecture
========================

RetryRunner helps to control the rate of connections establishment by limiting the number of
connector workers.

For example, if ``num_connectors`` is 5, it means there are only 5 workers
to establish connections to devices, even if there are 100 devices, RetryRunner will connect
only with 5 of them at a time. This is very helpful when connections rate need to be limited
due to operations restrictions like AAA (TACACS, RADIUS) servers load.

When new task started and if no connection exist to device that this task makes the use of,
RetryRunner attempts to connect to device retrying up to ``connect_retry`` times.

Once connection established, task handed over to worker threads for execution, workers
will retry up to ``task_retry`` if task fails.

Connection parameters such as timeouts or usage of SSH keys handled by Nornir Connection plugins.
RetryRunner calls Nornir to start the connection, further connection establishment details
controlled by Connection plugin itself.

.. image:: ../_images/RetryRunner_v0.png

RetryRunner Sample Usage
========================

Need to instruct Nornir to use RetryRunner on instantiation::

    from nornir import InitNornir

    NornirObj = InitNornir(
        runner={
            "plugin": "RetryRunner",
            "options": {
                "num_workers": 100,
                "num_connectors": 10,
                "connect_retry": 3,
                "connect_backoff": 1000,
                "connect_splay": 100,
                "task_retry": 3,
                "task_backoff": 1000,
                "task_splay": 100
            }
        }
    )

Sample code to demonstrate usage of ``RetryRunner``, ``DictInventory`` and ``ResultSerializer`` plugins::

    import yaml
    import pprint
    from nornir import InitNornir
    from nornir.core.task import Result, Task
    from nornir_netmiko import netmiko_send_command, netmiko_send_config
    from nornir_salt.plugins.functions import ResultSerializer

    inventory_data = '''
    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        groups: [lab]
      R2:
        hostname: 192.168.1.153
        platform: ios
        groups: [lab]
      R3:
        hostname: 192.168.1.154
        platform: ios
        groups: [lab]

    groups:
      lab:
        username: cisco
        password: cisco
    '''

    inventory_dict = yaml.safe_load(inventory_data)

    NornirObj = InitNornir(
        runner={
            "plugin": "RetryRunner",
            "options": {
                "num_workers": 100,
                "num_connectors": 10,
                "connect_retry": 3,
                "connect_backoff": 1000,
                "connect_splay": 100,
                "task_retry": 3,
                "task_backoff": 1000,
                "task_splay": 100
            }
        },
        inventory={
            "plugin": "DictInventory",
            "options": {
                "hosts": inventory_dict["hosts"],
                "groups": inventory_dict["groups"],
                "defaults": inventory_dict.get("defaults", {})
            }
        },
    )

    def _task_group_netmiko_send_commands(task, commands):
        # run commands
        for command in commands:
            task.run(
                task=netmiko_send_command,
                command_string=command,
                name=command
            )
        return Result(host=task.host)

    # run single task
    result1 = NornirObj.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

    # run grouped tasks
    result2 = NornirObj.run(
        task=_task_group_netmiko_send_commands,
        commands=["show clock", "show run | inc hostname"],
        connection_name="netmiko"
    )

    # run another single task
    result3 = NornirObj.run(
        task=netmiko_send_command,
        command_string="show run | inc hostname"
    )

    NornirObj.close_connections()

    # Print results
    formed_result1 = ResultSerializer(result1, add_details=True)
    pprint.pprint(formed_result1, width=100)

    formed_result2 = ResultSerializer(result2, add_details=True)
    pprint.pprint(formed_result2, width=100)

    formed_result3 = ResultSerializer(result3, add_details=True)
    pprint.pprint(formed_result3, width=100)

RetryRunner - Connect to hosts behind jumphost
==============================================

RetryRunner implements logic to connect with hosts behind bastion/jumphosts.

To connect to devices behind jumphost, need to define jumphost parameters in host's inventory data::

    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        username: test
        password: test
        data:
          jumphost:
            hostname: 10.1.1.1
            port: 22
            password: jump_host_password
            username: jump_host_user

.. note:: Only Netmiko ``connection_name="netmiko"`` and Ncclient ``connection_name="ncclient"``
    tasks, support connecting to hosts behind Jumphosts using above inventory data.

RetryRunner Reference
=====================

.. autoclass:: nornir_salt.plugins.runners.RetryRunner.RetryRunner
"""
import threading
import queue
import logging
import time
import random
from typing import List
from nornir.core.task import AggregatedResult, Task, MultiResult, Result
from nornir.core.inventory import Host

try:
    import paramiko

    HAS_PARAMIKO = True
except ImportError:
    HAS_PARAMIKO = False

log = logging.getLogger(__name__)
LOCK = threading.Lock()


def worker(
    stop_event,
    connectors_q,
    work_q,
    task_backoff: int,
    task_splay: int,
    task_retry: int,
    connect_retry: int,
    reconnect_on_fail: bool,
):
    while not stop_event.is_set():
        try:
            work = work_q.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        task, host, params, result = work
        # check if has exception recorded in params by connector thread
        if params.get("connect_exception"):
            work_result = MultiResult(task.name)
            work_result.append(
                Result(
                    host,
                    result=params["connect_exception"],
                    failed=True,
                    exception=params["connect_exception"],
                    name=task.name,
                )
            )
        else:
            # check if need backoff task retry for this host
            if params["task_retry"] > 0:
                elapsed = time.time() - params["timestamp"]
                should_wait = (params["task_retry"] * task_backoff) / 1000
                if elapsed < should_wait:
                    work_q.put(work)
                    work_q.task_done()
                    continue
            log.info(
                "nornir_salt:RetryRunner {} - running task '{}'".format(
                    host.name, task.name
                )
            )
            time.sleep(random.randrange(0, task_splay) / 1000)  # nosec
            work_result = task.start(host)
        # check if need to run task retry logic
        if task.results.failed:
            log.error(
                "nornir_salt:RetryRunner {} - task execution retry attempt {} failed: '{}'".format(
                    host.name, params["task_retry"], work_result[0].exception
                )
            )
            if (
                params["task_retry"] < task_retry
                and params["connection_retry"] < connect_retry
            ):
                params["task_retry"] += 1
                params["timestamp"] = time.time()
                # recover task results for not to count task as failed
                for r in task.results:
                    r.failed = False
                    r.exception = None
                    r.skip_results = (
                        True  # for ResultSerializer to skip failed attempts
                    )
                if reconnect_on_fail:
                    # close host connections to retry them
                    close_host_connection(host, params["connection_name"])
                    connectors_q.put(work)
                    params["connection_retry"] += 1
                else:
                    work_q.put(work)
                work_q.task_done()
                continue
        # enreach result objects with runner statistics
        for result_item in work_result:
            result_item.connection_retry = params["connection_retry"]
            result_item.task_retry = params["task_retry"]
        with LOCK:
            result[host.name] = work_result
            del work_result
        work_q.task_done()
        log.info(
            "nornir_salt:RetryRunner {} - task '{}' completed".format(
                host.name, task.name
            )
        )


def connector(
    stop_event,
    connectors_q,
    work_q,
    connect_backoff: int,
    connect_splay: int,
    connect_retry: int,
    jumphosts_connections,
):
    while not stop_event.is_set():
        try:
            connection = connectors_q.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        task, host, params, result = connection
        # check if need back off connection retry for this host
        if params["connection_retry"] > 0:
            elapsed = time.time() - params["timestamp"]
            should_wait = (params["connection_retry"] * connect_backoff) / 1000
            if elapsed < should_wait:
                connectors_q.put(connection)
                connectors_q.task_done()
                continue
        # initiate connection to host
        connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
        connection_name = task.params.pop("connection_name", connection_name)
        # on connect retry get connection name from params
        connection_name = params.get("connection_name", connection_name)
        params.setdefault("connection_name", connection_name)
        if connection_name and connection_name not in host.connections:
            try:
                time.sleep(random.randrange(0, connect_splay) / 1000)  # nosec
                if host.get("jumphost") and connection_name in ["netmiko", "ncclient"]:
                    extras = host.get_connection_parameters(connection_name).extras
                    extras["sock"] = connect_to_device_behind_jumphost(
                        host, jumphosts_connections
                    )
                    host.open_connection(
                        connection_name, configuration=task.nornir.config, extras=extras
                    )
                else:
                    host.open_connection(
                        connection_name, configuration=task.nornir.config
                    )
                log.info(
                    "nornir_salt:RetryRunner {} - started connection: '{}'".format(
                        host.name, connection_name
                    )
                )
            except Exception as e:
                # close host connections to retry them
                close_host_connection(host, connection_name)
                err_msg = "nornir_salt:RetryRunner {} - connection retry attempt {}, error: '{}'".format(
                    host.name, params["connection_retry"], e
                )
                log.error(err_msg)
                if params["connection_retry"] < connect_retry:
                    params["connection_retry"] += 1
                    params["timestamp"] = time.time()
                    connectors_q.put(connection)
                    connectors_q.task_done()
                    continue
                else:
                    # record exception in params for worker thread to react on it
                    params["connect_exception"] = err_msg
        connectors_q.task_done()
        work_q.put((task, host, params, result))


def close_host_connection(host, connection_name):
    try:
        host.close_connection(connection_name)
    except:
        _ = host.connections.pop(connection_name, None)
    if host.get("jumphost"):
        channel_name = "jumphost_{}_channel".format(host["jumphost"]["hostname"])
        try:
            host.close_connection(channel_name)
        except:
            _ = host.connections.pop(channel_name, None)


def connect_to_device_behind_jumphost(host, jumphosts_connections):
    """
    Establish connection to devices behind jumphost/bastion
    """

    jumphost = {"timeout": 3, "look_for_keys": False, "allow_agent": False}
    jumphost.update(host["jumphost"])
    # Initiate connection to jumphost if not initiated already or
    # if it failed before
    #
    # jumphosts_connections dictionary shared between threads,
    # first thread to start will detect that no connection to
    # jumphost exists and will try to establish it, while doing
    # so, jumphost key will be added to jumphosts_connections
    # dictionary with '__connecting__' value, next threads
    # will see '__connecting__' as value for jumphost and will
    # continue sleeping until either connection succeeded or
    # failed. On connection failure, first thread will set value
    # to '__failed__' to signal other threads to exit.
    if (
        jumphost["hostname"] not in jumphosts_connections
        or jumphosts_connections.get(jumphost["hostname"]) == "__failed__"
    ):
        try:
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = "__connecting__"
            jumphost_ssh_client = paramiko.client.SSHClient()
            jumphost_ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            jumphost_ssh_client.connect(**jumphost)
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = {
                    "jumphost_ssh_client": jumphost_ssh_client,
                    "jumphost_ssh_transport": jumphost_ssh_client.get_transport(),
                }
            # add jumphost to host connections to close it
            # on nornir.close_connections() call
            host.connections[
                "jumphost_{}".format(jumphost["hostname"])
            ] = jumphost_ssh_client
            log.info(
                "nornir_salt:RetryRunner Started connection to jumphost '{}' - '{}'".format(
                    jumphost["hostname"], jumphost_ssh_client
                )
            )
        except Exception as e:
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = "__failed__"
            # add exception info to host data to include in results
            error_msg = "nornir_salt:RetryRunner failed connection to jumphost '{}', error - {}".format(
                host["jumphost"]["hostname"], e
            )
            # host["exception"] = error_msg
            log.error(error_msg)
            raise RuntimeError(error_msg)
    else:
        # sleep random time waiting for connection to jumphost to establish
        while jumphosts_connections[jumphost["hostname"]] == "__connecting__":
            time.sleep(random.randrange(0, 500) / 1000)  # nosec
        if jumphosts_connections[jumphost["hostname"]] == "__failed__":
            # add exception info to host data to include in results
            error_msg = "nornir_salt:RetryRunner failed connection to jumphost '{}' in another thread".format(
                host["jumphost"]["hostname"]
            )
            # host["exception"] = error_msg
            log.error(error_msg)
            raise RuntimeError(error_msg)
    # connect to host
    channel_name = "jumphost_{}_channel".format(jumphost["hostname"])
    jconn = jumphosts_connections[jumphost["hostname"]]
    # check if connection to jumphost was closed, reconnect if so
    if not jconn["jumphost_ssh_transport"].is_active():
        log.debug(
            "nornir_salt:RetryRunner jumphost '{}' disconnected, reconnecting".format(
                jumphost["hostname"]
            )
        )
        jconn["jumphost_ssh_client"].close()
        jumphosts_connections.pop(jumphost["hostname"])
        return connect_to_device_behind_jumphost(host, jumphosts_connections)
    # proceed with openning new channel via jumphost
    if not host.connections.get(channel_name):
        host.connections[channel_name] = jconn["jumphost_ssh_transport"].open_channel(
            kind="direct-tcpip",
            dest_addr=(host.hostname, host.port or 22),
            src_addr=("localhost", 7777),
            timeout=3,
        )
        log.info(
            "nornir_salt:RetryRunner {} - started new channel via jumphost '{}' - '{}'".format(
                host.name, jumphost["hostname"], jconn["jumphost_ssh_client"]
            )
        )
    return host.connections[channel_name]


class RetryRunner:
    """
    RetryRunner is a modification of QueueRunner that strives to
    make task execution as reliable as possible.

    Arguments:
        num_workers: number of threads for tasks execution
        num_connectors: number of threads for device connections
        connect_retry: number of connection attempts
        connect_backoff: exponential backoff timer in milliseconds
        connect_splay: random interval between 0 and splay for each connection in milliseconds
        task_retry: number of attempts to run task
        task_backoff: exponential backoff timer in milliseconds
        task_splay: random interval between 0 and splay before task start in milliseconds
        reconnect_on_fail: boolean, default True, perform reconnect to host on task failure
        task_timeout: int, seconds to wait for task to complete before closing all queues
            and stopping connectors and workers threads, default 600
    """

    def __init__(
        self,
        num_workers: int = 100,
        num_connectors: int = 20,
        connect_retry: int = 3,
        connect_backoff: int = 5000,
        connect_splay: int = 100,
        task_retry: int = 1,
        task_backoff: int = 5000,
        task_splay: int = 100,
        reconnect_on_fail: bool = True,
        task_timeout: int = 600,
    ) -> None:
        self.num_workers = num_workers
        self.num_connectors = num_connectors
        self.connect_retry = connect_retry
        self.connect_backoff = connect_backoff
        self.connect_splay = connect_splay
        self.task_retry = task_retry
        self.task_backoff = task_backoff
        self.task_splay = task_splay
        self.jumphosts_connections = {}
        self.reconnect_on_fail = reconnect_on_fail
        self.task_timeout = task_timeout

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        connectors_q = queue.Queue()
        work_q = queue.Queue()
        stop_event = threading.Event()
        task_timeout_event = threading.Event()
        task_timeout_timer = threading.Timer(
            self.task_timeout, lambda: task_timeout_event.set()
        )
        result = AggregatedResult(task.name)
        connector_threads = []
        worker_threads = []
        # enqueue hosts in connectors queue
        for host in hosts:
            connectors_q.put(
                (task.copy(), host, {"connection_retry": 0, "task_retry": 0}, result)
            )
        # start connectors threads
        for _ in range(self.num_connectors):
            t = threading.Thread(
                target=connector,
                args=(
                    stop_event,
                    connectors_q,
                    work_q,
                    self.connect_backoff,
                    self.connect_splay,
                    self.connect_retry,
                    self.jumphosts_connections,
                ),
            )
            t.start()
            connector_threads.append(t)
        # start worker threads
        for _ in range(self.num_workers):
            t = threading.Thread(
                target=worker,
                args=(
                    stop_event,
                    connectors_q,
                    work_q,
                    self.task_backoff,
                    self.task_splay,
                    self.task_retry,
                    self.connect_retry,
                    self.reconnect_on_fail,
                ),
            )
            t.start()
            worker_threads.append(t)
        # wait until all hosts completed task or timeout reached
        task_timeout_timer.start()
        while not task_timeout_event.is_set():
            with LOCK:
                hosts_no_result = [h.name for h in hosts if h.name not in result]
            if hosts_no_result == []:
                task_timeout_timer.cancel()
                break
            time.sleep(0.1)
        else:
            log.warning(
                "nornir_salt:RetryRunner task '{}', '{}'s task_timeout reached, hosts no results '{}'".format(
                    task.name, self.task_timeout, hosts_no_result
                )
            )
        # block until all queues empty
        connectors_q.join()
        work_q.join()
        # stop and delete connectors and workers threads
        stop_event.set()
        while connector_threads:
            _ = connector_threads.pop().join()
        while worker_threads:
            _ = worker_threads.pop().join()
        return result



"""
scrapli_send_config
###################

This task plugin uses ``nornir-scrapli`` ``scrapli_send_config`` task
to send configuration commands to devices over SSH or Telnet.

``scrapli_send_config`` exists as part of ``nornir_salt`` repository to facilitate
per-host configuration rendering performed by SALT prior to running the task.

Dependencies:

* `nornir-scrapli module <https://pypi.org/project/nornir-scrapli/>`_ required

scrapli_send_config sample usage
================================

Code to invoke ``scrapli_send_config`` task::

    from nornir import InitNornir
    from nornir_salt import scrapli_send_config

    nr = InitNornir(config_file="config.yaml")

    output = nr.run(
        task=scrapli_send_config,
        commands=["sinterface loopback 0", "description 'configured by script'"]
    )

scrapli_send_config returns
===========================

Returns Nornir results object with individual tasks names set
equal to commands sent to device.

scrapli_send_config reference
=============================

.. autofunction:: nornir_salt.plugins.tasks.scrapli_send_config.scrapli_send_config
"""
import logging
from nornir.core.task import Result, Task

try:
    from nornir_scrapli.tasks import send_config as nornir_scrapli_send_config

    HAS_SCRAPLI = True
except ImportError:
    HAS_SCRAPLI = False

log = logging.getLogger(__name__)


def scrapli_send_config(task: Task, config=None, **kwargs):
    """
    Nornir Task function to send confgiuration to devices using
    ``nornir_scrapli.tasks.send_config`` plugin

    :param kwargs: arguments for ``file.apply_template_on_contents`` salt function
        for configuration rendering as well as for ``task.run`` method
    :param config: (str or list) configuration string or list of commands to send to device
    :param commit: not implemented yet
    :return result: Nornir result object with task execution results
    """
    # run sanity check
    if not HAS_SCRAPLI:
        return Result(
            host=task.host,
            failed=True,
            exception="No nornir_scrapli found, is it installed?",
        )

    # get configuration
    if "config" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["config"]
    elif "commands" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["filename"]

    # transform configuration to string if list/tuple given
    if isinstance(config, (list, tuple)):
        config = "\n".join(config)

    # push config to device
    task.run(
        task=nornir_scrapli_send_config,
        config=config,
        name="scrapli_send_config",
        **kwargs
    )

    # set skip_results to True, for ResultSerializer to ignore
    # results for grouped task itself, which are usually None
    return Result(host=task.host, skip_results=True)



from .ResultSerializer import ResultSerializer
from .FFun import FFun, FFun_functions
from .HostsKeepalive import HostsKeepalive
from .TabulateFormatter import TabulateFormatter
from .DumpResults import DumpResults
from .InventoryFun import InventoryFun

__all__ = (
    "ResultSerializer",
    "FFun",
    "HostsKeepalive",
    "TabulateFormatter",
    "DumpResults",
    "InventoryFun",
    "FFun_functions",
)



"""
TabulateFormatter
#################

Function to transform results in a text table format using Tabulate module.

TabulateFormatter works with a list of dictionaries to represent them as a table,
if Nornir AggregatedResult object passed on to TabulateFormatter it uses
ResultSerializer function to serialize results into a list of dictionaries.

Dependencies:

* Nornir 3.x.x
* `Tabulate module <https://pypi.org/project/tabulate/>`_ for results table formatting

Sample code to use TabulateFormatter::

    from nornir import InitNornir
    from nornir_salt import TabulateFormatter
    from nornir_netmiko import netmiko_send_command

    nr = InitNornir(config_file="config.yaml")

    result = NornirObj.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

    print(TabulateFormatter(result))
    # prints:
    # result                    changed    diff    failed    name          connection_retry    task_retry  exception    host
    # ------------------------  ---------  ------  --------  ----------  ------------------  ------------  -----------  ------
    # Sun Jul 11 08:41:06 2021  False              False     show clock                   0             0               ceos1
    # Timezone: UTC
    # Clock source: local
    # Sun Jul 11 08:41:06 2021  False              False     show clock                   0             0               ceos2
    # Timezone: UTC
    # Clock source: local

Reference
=========

.. autofunction:: nornir_salt.plugins.functions.TabulateFormatter.TabulateFormatter
"""
import logging
from nornir.core.task import AggregatedResult
from .ResultSerializer import ResultSerializer

log = logging.getLogger(__name__)

try:
    import tabulate as tabulate_lib

    HAS_TABULATE = True
except ImportError:
    HAS_TABULATE = False
    log.error("Failed to import tabulate library, install it: pip install tabulate")


def TabulateFormatter(
    result,
    tabulate=True,
    headers="keys",
    headers_exclude=None,
    sortby=None,
    reverse=False,
):
    """
    Function to format results in a text table.

    :param result: list of dictionaries or ``nornir.core.task.AggregatedResult`` object
    :param tabulate: (dict or str or bool) controls tabulate behavior
    :param headers: (list or str) list of table headers, comma-separated string of headers or
        one of tabulate supported values, e.g. ``keys``
    :param headers_exclude: (list) list of table headers, comma-separated string of headers
        to exclude
    :param sortby: (str) name of the key to sort table by, default is ``None`` - no sorting applied
    :param reverse: (bool) reverses sort order if True, default is False

    Supported values for ``tabulate`` attribute:

    * ``brief`` - ``tablefmt`` is ``grid``, ``showindex`` is ``True``, ``headers`` are ``host, name, result, exception``
    * ``terse`` - ``tablefmt`` is ``simple``, ``showindex`` is ``True``, ``headers`` are ``host, name, result, exception``
    * ``True`` - uses ``headers``, no other formatting
    * ``False`` - does nothing, returns original results
    * ``extend`` - if result is a list, extends it to form final table, appends it as is otherwise
    * ``dictionary`` - dictionary content passed as ``**kwargs`` to ``tabulate.tabulate`` method
    """
    if not HAS_TABULATE:
        log.error(
            "nornir-salt:TabulateFormatter failed import tabulate library, install: pip install tabulate"
        )
        return result

    headers_exclude = headers_exclude or []

    # decide on results to tabulate
    if isinstance(result, AggregatedResult):
        result_to_tabulate = ResultSerializer(result, add_details=True, to_dict=False)
    elif isinstance(result, list):
        result_to_tabulate = result
    else:
        log.error(
            "nornir-salt:TabulateFormatter unsupported results type '{}', supported - list or AggregatedResult".format(
                type(result)
            )
        )
        return result

    # check headers
    if isinstance(headers, str) and "," in headers:
        headers = [i.strip() for i in headers.split(",")]
    if isinstance(headers_exclude, str) and "," in headers_exclude:
        headers_exclude = [i.strip() for i in headers_exclude.split(",")]

    # form tabulate parameters and results
    if tabulate in ["brief", "terse"]:
        tabulate = {
            "tablefmt": "grid" if tabulate == "brief" else "simple",
            "showindex": True,
            "headers": ["host", "name", "result", "exception"],
        }
    elif tabulate is True:
        tabulate = {"headers": headers}
    elif tabulate == "extend":
        table_ = []
        tabulate = {"headers": headers}
        for res in result_to_tabulate:
            if isinstance(res["result"], list):
                table_.extend(
                    [
                        {**res, **i} if isinstance(i, dict) else {**res, "result": i}
                        for i in res.pop("result")
                    ]
                )
            else:
                table_.append(res)
        result_to_tabulate = table_
    elif isinstance(tabulate, dict):
        tabulate.setdefault("headers", headers)
    elif tabulate is False:
        return ResultSerializer(result, add_details=True, to_dict=False)
    else:
        log.error(
            "nornir-salt:TabulateFormatter unsupported tabulate argument type '{}', value '{}', supported - 'brief', bool, dict".format(
                type(tabulate), str(tabulate)
            )
        )
        return ResultSerializer(result, add_details=True, to_dict=False)

    # sort results
    if sortby and isinstance(sortby, str):
        result_to_tabulate = sorted(
            result_to_tabulate,
            reverse=reverse,
            key=lambda item: str(item.get(sortby, "")),
        )

    # filter table headers if requested to do so
    if headers_exclude:
        result_to_tabulate = [
            {k: v for k, v in res.items() if k not in headers_exclude}
            for res in result_to_tabulate
        ]

    # transform result_to_tabulate to list of lists
    if isinstance(tabulate["headers"], list):
        result_to_tabulate = [
            [item.get(i, "") for i in tabulate["headers"]]
            for item in result_to_tabulate
        ]

    return tabulate_lib.tabulate(result_to_tabulate, **tabulate)



"""
netmiko_send_commands
#####################

This task plugin uses ``nornir-netmiko`` ``netmiko_send_command`` task
to send multiple commands to devices.

Dependencies:

* `nornir-netmiko module <https://pypi.org/project/nornir-netmiko/>`_ required

netmiko_send_commands sample usage
==================================

Code to invoke ``netmiko_send_commands`` task::

    from nornir_salt import netmiko_send_commands

    output = nr.run(
        task=netmiko_send_commands,
        commands=["show run", "show clock"]
    )

netmiko_send_commands returns
=============================

Returns Nornir results object with individual tasks names set
equal to commands sent to device.

netmiko_send_commands reference
===============================

.. autofunction:: nornir_salt.plugins.tasks.netmiko_send_commands.netmiko_send_commands
"""
import time
import logging
from nornir.core.task import Result, Task
from .netmiko_send_command_ps import netmiko_send_command_ps

try:
    from nornir_netmiko.tasks import netmiko_send_command

    HAS_NETMIKO = True
except ImportError:
    HAS_NETMIKO = False

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "netmiko"


def netmiko_send_commands(
    task: Task,
    commands=None,
    interval=0.01,
    use_ps: bool = False,
    split_lines: bool = True,
    new_line_char: str = "_br_",
    **kwargs
):
    """
    Nornir Task function to send show commands to devices using
    ``nornir_netmiko.tasks.netmiko_send_command`` plugin

    Per-host ``commands`` can be provided using host's object ``data`` attribute
    with ``__task__`` key with value set to dictionary with ``commands`` key
    containing a list of or a multiline string of commands to send to device, e.g.::

        print(host.data["__task__"]["commands"])

        ["ping 1.1.1.1 source 1.1.1.2", "show clock"]

    Alternatively, ``__task__`` can contain ``filename`` key with commands string
    to send to device.

    :param kwargs: (dict) any additional arguments to pass to ``netmiko_send_command``
        ``nornir-netmiko`` task
    :param commands: (list or str) list or multiline string of commands to send to device
    :param interval: (int) interval between sending commands, default 0.01s
    :param use_ps: (bool) set to True to switch to experimental send_command_ps method
    :param split_lines: (bool) if True (default) - split multiline string to commands,
        if False, send multiline string to device as is; honored only when ``use_ps`` is
        True, ``split_lines`` ignored if ``use_ps`` is False
    :param new_line_char: (str) characters to replace in commands with new line ``\\n``
        before sending command to device, default is ``_br_``, useful to simulate enter key
    :return result: Nornir result object with task results named after commands
    """
    # run sanity check
    if not HAS_NETMIKO:
        return Result(
            host=task.host,
            failed=True,
            exception="No nornir_netmiko found, is it installed?",
        )

    commands = commands or []

    # run interval sanity check
    interval = interval if isinstance(interval, (int, float)) else 0.01

    # get per-host commands if any
    if "commands" in task.host.data.get("__task__", {}):
        if commands:
            for c in task.host.data["__task__"]["commands"]:
                if c not in commands:
                    commands.append(c)
        else:
            commands = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        commands = task.host.data["__task__"]["filename"]

    # normalize commands to a list
    if isinstance(commands, str) and split_lines:
        commands = commands.splitlines()
    elif isinstance(commands, str) and not split_lines:
        commands = [commands]

    # remove empty lines/commands that can left after rendering
    commands = [c for c in commands if c.strip()]

    # iterate over commands and see if need to add empty line - hit enter
    commands = [
        c.replace(new_line_char, "\n") if new_line_char in c else c for c in commands
    ]

    # run commands
    if use_ps:
        # send commands
        for index, command in enumerate(commands):
            task.run(
                task=netmiko_send_command_ps,
                command_string=command,
                name=command.strip().splitlines()[0],
                **kwargs
            )
            # do not sleep after last command sent
            if index != len(commands) - 1:
                time.sleep(interval)
    else:
        # send commands
        for command in commands:
            task.run(
                task=netmiko_send_command,
                command_string=command,
                name=command.strip(),
                **kwargs
            )
            time.sleep(interval)

    # set skip_results to True, for ResultSerializer to ignore
    # results for grouped task itself, which are usually None
    return Result(host=task.host, skip_results=True)



"""
nr_test
#######

Function to test Nornir operation and produce valid results
without the need to interact with devices directly.

nr_test sample usage
====================

Code to invoke ``nr_test`` task::

    from nornir_salt import nr_test

    output = nr.run(task=nr_test, abc=123)

nr_test returns
=============================

Returns Nornir results object with task results populated
in accordance with arguments supplied to task call.

nr_test reference
=================

.. autofunction:: nornir_salt.plugins.tasks.nr_test.nr_test
"""
from nornir.core.task import Result


def nr_test(
    task,
    ret_data_per_host=None,
    ret_data="__undefined_value__",
    excpt=None,
    excpt_msg="",
    **kwargs
):
    """
    Dummy task that echoes data passed to it. Useful to debug and
    verification of Nornir object operation.

    :param ret_data_per_host: (dict) Dictionary keyed by host name with
        values to include in results
    :param ret_data: Any data to include in results, same for each host
    :param kwargs: Any key-value pair to include in results, same for each host
    :return result: ``ret_data`` or ``**kwargs`` passed to the task
    :param excpt: (obj or True) exception object to raise; if True, raises ``RuntimeError``
    :param excpt_msg: (str) message to use with exception

    Order of preference of return data:

    1. If ``ret_data_per_host`` present, it is used to form results
    2. If ``excpt`` object supplied, it is raised
    3. If ``ret_data`` present, it is included in results
    4. If ``**kwargs`` supplied, they are included in results
    """
    ret_data_per_host = ret_data_per_host or {}

    if ret_data_per_host:
        return Result(
            host=task.host, result=ret_data_per_host.get(task.host.name, None)
        )
    elif excpt is True:
        raise RuntimeError(excpt_msg)
    elif excpt is not None:
        raise excpt(excpt_msg)
    elif ret_data != "__undefined_value__":
        return Result(host=task.host, result=ret_data)
    else:
        return Result(host=task.host, result=kwargs)



"""
pyats_genie_api
###############

This task can call one of Genie device ``api`` methods.

Full list of api methods available for your device platform can be found
`in Genie docs <https://pubhub.devnetcloud.com/media/genie-feature-browser/docs/#/apis>`_

Dependencies:

* `PyATS library <https://pypi.org/project/pyats/>`_ required
* `Genie library <https://pypi.org/project/genie/>`_ required

Sample Usage
============

Code to invoke ``pyats_genie_api`` task::

    from nornir_salt import pyats_genie_api

    output = nr.run(
        task=pyats_genie_api,
        api="ping",
        address="127.0.0.1",
        timeout=1,
        count=1,
    )

``pyats_genie_api`` returns Nornir results object with results containing
api call output.

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.pyats_genie_api.pyats_genie_api
"""
import logging
from nornir.core.task import Result, Task

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "pyats"


def pyats_genie_api(task: Task, api: str, **kwargs):
    """
    Salt-nornir task function to call Genie device's api methods.

    :param api: (str) api name to call
    :param kwargs: (dict) any additional arguments to use with api call
    :return result: Nornir result object with task execution results
    """
    task.name = "pyats_genie_api"

    # get PyATS testbed, device object
    testbed = task.host.get_connection(CONNECTION_NAME, task.nornir.config)
    device = testbed.devices[task.host.name]

    # call api
    result = getattr(device.api, api)(**kwargs)

    return Result(host=task.host, result=result)



"""
FFun
####

Nornir interacts with many devices and has it's own inventory. FFun is an extension
that uses Nornir's built-in filtering capabilities to narrow down tasks execution
to certain hosts/devices.

Filtering order::

    FO -> FB -> FC -> FR -> FG -> FP -> FL -> FM -> FN

.. note:: If multiple filters provided, hosts must pass all checks - ``AND`` logic - to succeed.


FFun filters overview
=====================

FO - Filter Object
------------------

Filter using `Nornir Filter Object <https://nornir.readthedocs.io/en/latest/tutorial/inventory.html#Filter-Object>`_

Examples::

    # Platform ios and hostname 192.168.217.7:
    filtered_hosts = FFun(NornirObj, FO={"platform": "ios", "hostname": "192.168.217.7"})

    # Host with name R1 or R2:
    filtered_hosts = FFun(NornirObj, FO={"name__any": ["R1", "R2"])

    # Location B1 or location B2:
    filtered_hosts = FFun(NornirObj, FO=[{"location": "B1"}, {"location": "B2"}])

    # Location B1 and platform ios or any host at location B2:
    filtered_hosts = FFun(NornirObj, FO=[{"location": "B1", "platform": "ios"}, {"location": "B2"}])

FB - Filter gloB
----------------

Filter hosts by name using Glob Patterns matching `fnmatchcase <https://docs.python.org/3.4/library/fnmatch.html#fnmatch.fnmatchcase>`_ module::

    # Match R1, R2, R# hostnames but not R11 or R4:
    filtered_hosts = FFun(NornirObj, FB="R[123]")

    # Match R1, R2, and SW1 but not R11 or R4 or eSW1 using list of patterns:
    filtered_hosts = FFun(NornirObj, FB=["R[12]", "SW*"])

    # Match R1, R2, and SW1 but not R11 or R4 or eSW1 using comma separated list of patterns:
    filtered_hosts = FFun(NornirObj, FB="R[12], SW*")

If list of patterns provided, host matching at least one pattern will pass this check.

FC - Filter Contains Any
------------------------

Filter hosts by checking if their name contains any of the string patterns::

    # Match core-switch-1 but not switch-1:
    filtered_hosts = FFun(NornirObj, FC="core-switch")

    # Match R1, R2, and SW1 but not ER33 or CR4 using list of patterns:
    filtered_hosts = FFun(NornirObj, FC=["R1", "R2", "SW"])

    # Match R1, R2, and SW1 but not ER33 or CR4 using comma separated list of patterns:
    filtered_hosts = FFun(NornirObj, FC="R1, R2, SW")

If list of patterns provided, host matching at least one pattern will pass this check.

FR - Filter Regex
-----------------

Filter hosts by checking if their name contains any of regular expression patterns::

    # Match core-switch-1 but not switch-1:
    filtered_hosts = FFun(NornirObj, FR=".+core-switch.+")

    # Match R1, R2, and SW1 but not ER33 or CR4 using list of patterns:
    filtered_hosts = FFun(NornirObj, FR=["^R1$", "^R2$", "^SW$"])

If list of patterns provided, host matching at least one pattern will pass this check.

FG - Filter Group
-----------------

Filter hosts by group returning all hosts that belongs to given group::

    # return hosts that belong to 'lab' group only
    filtered_hosts = FFun(NornirObj, FG="lab")

FP - Filter Prefix
------------------

Filter hosts by checking if hosts hostname is part of at least one of given IP Prefixes::

    # return hosts with hostnames in 192.168.217.0/29, 192.168.2.0/24 ranges
    filtered_hosts = FFun(NornirObj, FP="192.168.217.0/29, 192.168.2.0/24")

If host's inventory hostname is IP, will use it as is, if it is FQDN, will
attempt to resolve it to obtain IP address, if DNS resolution fails, host
fails the check.

FL - Filter List
----------------

Match only hosts with names in provided list::

    filtered_hosts = FFun(NornirObj, FL="R1, R2")

FM - Filter platforM
--------------------

Match only hosts with given platform name patterns::

    filtered_hosts = FFun(NornirObj, FM="cisco*, huawei*")
    filtered_hosts = FFun(NornirObj, FM="cisco*")
    filtered_hosts = FFun(NornirObj, FM=["cisco*", "huawei*"])

If list of patterns provided, host with platform matching at least one pattern passes this check.

FN - Filter Negate
------------------

Negate matching results if ``FN`` argument set to ``True``::

    # will match all hosts except R1 and R2
    filtered_hosts = FFun(NornirObj, FL="R1, R2", FN=True)

FFun passes through all the ``Fx`` functions filtering hosts normally, ``FN``
function called at the end to form a set of non matched hosts, that set used
with ``FL`` function to provide final match result.

FFun sample usage
=================

Example how to invoke FFun filtering::

    import pprint
    import yaml
    from nornir import InitNornir
    from nornir_netmiko import netmiko_send_command
    from nornir_salt.plugins.functions import FFun

    inventory_data = '''
    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        groups: [lab]
        data:
          role: core
          site: B1
      SW1:
        hostname: 192.168.2.144
        platform: nxos_ssh
        groups: [lab, pod1]
        data:
          role: access
          site: B3

    groups:
      lab:
        username: cisco
        password: cisco
      pod1:
        username: cisco@
        password: cisco
    '''

    inventory_dict = yaml.safe_load(inventory_data)

    NornirObj = InitNornir(config_file="config.yaml")

    filtered_hosts = FFun(NornirObj, FB="R*", FG="lab", FP="192.168.1.0/24", FO={"role": "core"})

    pprint.pprint(filtered_hosts.dict()["inventory"]["hosts"])

    # should print:
    # {'R1': {'connection_options': {},
    #         'data': {'role': 'core', 'site': 'B1'},
    #         'groups': ['lab'],
    #         'hostname': '192.168.1.151',
    #         'name': 'R1',
    #         'password': 'cisco',
    #         'platform': 'ios',
    #         'port': None,
    #         'username': 'cisco'}}

    result = filtered_hosts.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

FFun return
===========

Nornir object with filtered hosts to further execute tasks against.

FFun reference
==============

.. autofunction:: nornir_salt.plugins.functions.FFun.FFun
"""

import logging
from fnmatch import fnmatchcase
from nornir.core.filter import F


log = logging.getLogger(__name__)

# list that enumerates all available FFun functions, useful for arguments filtering
FFun_functions = ["FO", "FB", "FC", "FR", "FG", "FP", "FL", "FM", "FN"]


def FFun(nr, check_if_has_filter=False, **kwargs):
    """
    Inventory filters dispatcher function.

    :param nr: Nornir object
    :param kwargs: Dictionary with filtering parameters e.g. {"FB": "host1*", "FL": ["host1", "host2"]}
    :param check_if_has_filter: (bool) default is False, if True, returns tuple ``(filtered_hosts, has_filter)``,
        where ``has_filter`` is boolean set to True if any of ``Fx`` filters provided
    :param FO: (str) Nornir Filter object dictionary
    :param FB: (str or list) glob pattern or comma separate list of patterns to filter based on hosts' names
    :param FC: (str or list) pattern or comma separate list of patterns to check for containment in hostname
    :param FR: (str or list) regex pattern or list of patterns to filter based on hosts' names
    :param FG: (str) Name of inventory group to return only hosts that part of it
    :param FP: (str) string, comma separated list of IPv4 or IPv6 prefixes e.g. 102.168.1.0/24
    :param FL: (str) string, comma separated list of hosts' names to return
    :param FN: (bool) default is False, if True, will negate match results to opposite
        set of hosts
    """
    ret = nr
    has_filter = False
    # check if kwargs argument given, usually
    # supplied by SALT nornir-proxy, that required for pop to modify
    # Nornir proxy original kwargs by removing all Fx arguments
    if kwargs.get("kwargs"):
        kwargs = kwargs["kwargs"]
    if kwargs.get("FO"):
        ret = _filter_FO(ret, kwargs.pop("FO"))
        has_filter = True
    if kwargs.get("FB"):
        ret = _filter_FB(ret, kwargs.pop("FB"))
        has_filter = True
    if kwargs.get("FC"):
        ret = _filter_FC(ret, kwargs.pop("FC"))
        has_filter = True
    if kwargs.get("FR"):
        ret = _filter_FR(ret, kwargs.pop("FR"))
        has_filter = True
    if kwargs.get("FG"):
        ret = _filter_FG(ret, kwargs.pop("FG"))
        has_filter = True
    if kwargs.get("FP"):
        ret = _filter_FP(ret, kwargs.pop("FP"))
        has_filter = True
    if "FL" in kwargs:
        ret = _filter_FL(ret, kwargs.pop("FL"))
        has_filter = True
    if "FM" in kwargs:
        ret = _filter_FM(ret, kwargs.pop("FM"))
        has_filter = True
    if "FN" in kwargs:
        ret = _filter_FN(ret, nr, kwargs.pop("FN"))
    return (ret, has_filter) if check_if_has_filter else ret


def _filter_FO(nr, filter_data):
    """
    Function to filter hosts using Filter Object
    """
    ret = nr
    if isinstance(filter_data, dict):
        ret = ret.filter(F(**filter_data))
    elif isinstance(filter_data, list):
        ret = ret.filter(F(**filter_data[0]))
        for item in filter_data[1:]:
            filtered_hosts = nr.filter(F(**item))
            ret.inventory.hosts.update(filtered_hosts.inventory.hosts)
    return ret


def _filter_FB(ret, pattern):
    """
    Function to filter hosts by name using glob patterns
    """
    # check if comma separated list of patterns given
    if isinstance(pattern, str) and "," in pattern:
        pattern = [i.strip() for i in pattern.split(",")]
    # run filtering
    if isinstance(pattern, list):
        return ret.filter(
            filter_func=lambda h: any([fnmatchcase(h.name, str(p)) for p in pattern])
        )
    else:
        return ret.filter(filter_func=lambda h: fnmatchcase(h.name, str(pattern)))


def _filter_FC(ret, pattern):
    """
    Function to filter hosts by name using pattern containment check
    """
    # check if comma separated list of patterns given
    if isinstance(pattern, str) and "," in pattern:
        pattern = [i.strip() for i in pattern.split(",")]
    # run filtering
    if isinstance(pattern, list):
        return ret.filter(
            filter_func=lambda h: any([str(p) in h.name for p in pattern])
        )
    else:
        return ret.filter(filter_func=lambda h: str(pattern) in h.name)


def _filter_FR(ret, pattern):
    """
    Function to filter hosts by name using regex pattern search
    """
    import re

    # run filtering
    if isinstance(pattern, list):
        return ret.filter(
            filter_func=lambda h: any([re.search(str(p), h.name) for p in pattern])
        )
    else:
        return ret.filter(
            filter_func=lambda h: True if re.search(str(pattern), h.name) else False
        )


def _filter_FG(ret, group):
    """
    Function to filter hosts using Groups
    """
    return ret.filter(filter_func=lambda h: h.has_parent_group(group))


def _filter_FP(ret, pfx):
    """
    Function to filter hosts based on IP Prefixes
    """
    import ipaddress
    import socket

    socket.setdefaulttimeout(1)

    def _filter_net(host):
        # convert host's ip to ip address object
        try:
            ip_addr = ipaddress.ip_address(host.hostname)
        except ValueError:
            # try to resolve hostname using DNS
            try:
                ip_str = socket.gethostbyname(host.hostname)
                ip_addr = ipaddress.ip_address(ip_str)
            except Exception as e:
                log.error(
                    "FP failed to convert host IP '{}', error '{}'".format(host.name, e)
                )
                return False
        # run filtering
        for net in networks:
            if ip_addr in net:
                return True
        return False

    # make a list of network objects
    prefixes = [i.strip() for i in pfx.split(",")] if isinstance(pfx, str) else pfx
    networks = []
    for prefix in prefixes:
        try:
            networks.append(ipaddress.ip_network(prefix))
        except Exception as e:
            log.error("FP failed to convert prefix '{}', error '{}'".format(prefix, e))
    # filter hosts
    return ret.filter(filter_func=_filter_net)


def _filter_FL(ret, names_list):
    """
    Function to filter hosts names based on list of names
    """
    names_list = (
        [i.strip() for i in names_list.split(",")]
        if isinstance(names_list, str)
        else names_list
    )
    if "_all_" in names_list:
        return ret
    else:
        return ret.filter(filter_func=lambda h: h.name in names_list)


def _filter_FM(ret, pattern):
    """
    Function to filter hosts by platform glob patterns
    """
    # check if comma separated list of patterns given
    if isinstance(pattern, str) and "," in pattern:
        pattern = [i.strip() for i in pattern.split(",")]
    # run filtering
    if isinstance(pattern, list):
        return ret.filter(
            filter_func=lambda h: any(
                [fnmatchcase(h.platform, str(p)) for p in pattern]
            )
        )
    else:
        return ret.filter(filter_func=lambda h: fnmatchcase(h.platform, str(pattern)))


def _filter_FN(ret, nr, FN):
    """
    Function to negate hosts' match results
    """
    if FN is not True:
        return ret
    all_hosts = set(nr.inventory.hosts.keys())
    matched_hosts = set(ret.inventory.hosts.keys())
    FNed_hosts = all_hosts.difference(matched_hosts)
    return _filter_FL(nr, list(FNed_hosts))



"""
Overview
########

Nornir-salt package is a collection of plugins and functions to use together with Nornir framework.

Primary reason for developing nornir-salt is a need to extend Nornir functionality for the
sake of salt-nornir proxy minion module. However, all plugins included in nornir-salt can
be used standalone unless stated otherwise.

While it was possible to include all the plugins into salt-nornir proxy module, decision was made
that many of the plugins developed can be useful outside of salt-nornir proxy module scope.
"""
# misc functions
from .plugins.functions import ResultSerializer
from .plugins.functions import FFun, FFun_functions
from .plugins.functions import TabulateFormatter
from .plugins.functions import DumpResults
from .plugins.functions import InventoryFun

# inventory plugins
from .plugins.inventory import DictInventory

# runner plugins
from .plugins.runners import QueueRunner
from .plugins.runners import RetryRunner

# task plugins
from .plugins.tasks import tcp_ping
from .plugins.tasks import netmiko_send_commands
from .plugins.tasks import netmiko_send_command_ps
from .plugins.tasks import nr_test
from .plugins.tasks import salt_cfg_gen
from .plugins.tasks import scrapli_send_commands
from .plugins.tasks import netmiko_send_config
from .plugins.tasks import napalm_configure
from .plugins.tasks import scrapli_send_config
from .plugins.tasks import http_call
from .plugins.tasks import file_read, file_list, file_remove, file_diff, files
from .plugins.tasks import connections, conn_close, conn_list
from .plugins.tasks import ncclient_call
from .plugins.tasks import scrapli_netconf_call
from .plugins.tasks import pygnmi_call
from .plugins.tasks import salt_clear_hcache
from .plugins.tasks import napalm_send_commands
from .plugins.tasks import sleep
from .plugins.tasks import pyats_send_commands
from .plugins.tasks import pyats_send_config
from .plugins.tasks import pyats_genie_api

# connections plugins
from .plugins.connections import NcclientPlugin
from .plugins.connections import HTTPPlugin
from .plugins.connections import PyGNMIPlugin
from .plugins.connections import ConnectionsPool
from .plugins.connections import PyATSUnicon

# processors plugins
from .plugins.processors import ToFileProcessor
from .plugins.processors import TestsProcessor
from .plugins.processors import DiffProcessor
from .plugins.processors import DataProcessor
from .plugins.processors import SaltEventProcessor

__all__ = (
    "ResultSerializer",
    "FFun",
    "FFun_functions",
    "DictInventory",
    "QueueRunner",
    "RetryRunner",
    "tcp_ping",
    "netmiko_send_commands",
    "netmiko_send_command_ps",
    "nr_test",
    "NcclientPlugin",
    "ncclient_call",
    "scrapli_netconf_call",
    "ToFileProcessor",
    "TestsProcessor",
    "TabulateFormatter",
    "DiffProcessor",
    "salt_cfg_gen",
    "scrapli_send_commands",
    "netmiko_send_config",
    "napalm_configure",
    "scrapli_send_config",
    "HTTPPlugin",
    "http_call",
    "file_read",
    "file_list",
    "file_remove",
    "file_diff",
    "files",
    "DumpResults",
    "connections",
    "conn_close",
    "conn_list",
    "PyGNMIPlugin",
    "pygnmi_call",
    "DataProcessor",
    "salt_clear_hcache",
    "napalm_send_commands",
    "ConnectionsPool",
    "sleep",
    "SaltEventProcessor",
    "PyATSUnicon",
    "pyats_send_commands",
    "pyats_send_config",
    "InventoryFun",
    "pyats_genie_api",
)



"""
napalm_send_commands
####################

This task plugin uses ``nornir-napalm`` ``napalm_cli`` task
to send multiple commands to devices pre-processing commands accordingly.

Pre-processing includes:

- Check and if any - retrieve per-host commands from host's inventory data
  ``task.host.data["__task__"]["commands"]`` or from ``task.host.data["__task__"]["filename"]``
- If command is a multi-line string, split it to individual lines or form a list with single command
- Iterate over commands list and remove empty strings
- Iterate over commands and replace ``new_line_char`` with ``\\n`` new line

Next, if ``interval`` argument provided commands send one by one to device using
``napalm_cli`` task plugin sleeping for given ``interval`` between commands. If
``interval`` argument is not provided, all commands sent at once.

Dependencies:

* `nornir-napalm module <https://pypi.org/project/nornir-napalm/>`_ required

napalm_send_commands sample usage
=================================

Code to invoke ``napalm_send_commands`` task::

    from nornir_salt import napalm_send_commands

    output = nr.run(
        task=napalm_send_commands,
        commands=["show run", "show clock"]
    )

Task ``napalm_send_commands`` returns Nornir results object with individual tasks
names set equal to commands sent to device.

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.napalm_send_commands.napalm_send_commands
"""
import time
import logging
from nornir.core.task import Result, Task

try:
    from nornir_napalm.plugins.tasks import napalm_cli

    HAS_NAPALM = True
except ImportError:
    HAS_NAPALM = False

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "napalm"


def napalm_send_commands(
    task: Task, commands=None, interval=None, new_line_char: str = "_br_"
):
    """
    Nornir Task function to send show commands to devices using ``napalm_cli`` task
    plugin.

    Per-host ``commands`` can be provided using host's object ``data`` attribute
    with ``__task__`` key with value set to dictionary with ``commands`` key
    containing a list of or a multiline string of commands to send to device, e.g.::

        print(host.data["__task__"]["commands"])

        ["ping 1.1.1.1 source 1.1.1.2", "show clock"]

    Alternatively, ``__task__`` can contain ``filename`` key with commands string
    to send to device.

    :param commands: (list or str) list or multiline string of commands to send to device
    :param interval: (int) interval between sending commands, default None
    :param new_line_char: (str) characters to replace in commands with new line ``\\n``
        before sending command to device, default is ``_br_``, useful to simulate enter key
    :return result: Nornir result object with task results named after commands
    """
    # run sanity check
    if not HAS_NAPALM:
        return Result(
            host=task.host,
            failed=True,
            exception="No nornir-napalm found, is it installed?",
        )

    commands = commands or []

    # get per-host commands if any
    if "commands" in task.host.data.get("__task__", {}):
        if commands:
            for c in task.host.data["__task__"]["commands"]:
                if c not in commands:
                    commands.append(c)
        else:
            commands = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        commands = task.host.data["__task__"]["filename"]

    # normalize commands to a list
    if isinstance(commands, str):
        commands = commands.splitlines()

    # remove empty lines/commands that can left after rendering
    commands = [c for c in commands if c.strip()]

    # iterate over commands and see if need to add empty line - hit enter
    commands = [
        c.replace(new_line_char, "\n") if new_line_char in c else c for c in commands
    ]

    # send commands one by one
    if isinstance(interval, (int, float)):
        for index, command in enumerate(commands):
            task.run(task=napalm_cli, commands=[command])
            # do not sleep after last command sent
            if index != len(commands) - 1:
                time.sleep(interval)
    # send all at once
    else:
        task.run(task=napalm_cli, commands=commands)

    # iterate over results and form per-command results
    per_command_results = []
    while task.results:
        res = task.results.pop()
        # check if task failed, do nothing if so
        if res.failed or res.exception:
            per_command_results.append(res)
            continue
        # iterate over dictionary result and construct per-command result
        for command, output in res.result.items():
            per_command_results.append(
                Result(host=task.host, result=output, name=command.strip())
            )
    task.results.extend(per_command_results)

    # set skip_results to True, for ResultSerializer to ignore
    # results for grouped task itself, which are usually None
    return Result(host=task.host, skip_results=True)



"""
salt_clear_hcache
#################

Task to clear hosts' task results cache stored  in data. This task
plugin does not have much practical applicability outside of
SaltStack environment.

salt_clear_hcache sample usage
==============================

Sample code to run ``salt_clear_hcache`` task::

    import pprint
    from nornir import InitNornir
    from nornir_salt.plugins.tasks import salt_clear_hcache

    nr = InitNornir(config_file="config.yaml")

    result = NornirObj.run(
        task=salt_clear_hcache,
        cache_keys=["cache_key1", "cache_key2"]
    )

    result_dictionary = ResultSerializer(result)

    pprint.pprint(result_dictionary)


salt_clear_cache returns dictionary of cleared cache keys with status

- True - cache key deleted
- False - cache key not found

salt_clear_cache reference
==========================

.. autofunction:: nornir_salt.plugins.tasks.salt_clear_hcache.salt_clear_hcache
"""
import logging
from typing import List

from nornir.core.task import Result, Task

log = logging.getLogger(__name__)


def salt_clear_hcache(task: Task, cache_keys: List = None, **kwargs) -> Result:
    """
    Function to iterate over provided cache keys and delete them from hosts's data.

    :param cache_keys: (list or str) list of cache keys to clean from host's data,
        if ``cache_keys`` argument not provided removes all cached data
    :returns: (dict) dictionary keyed by cache key and True/False status
    """
    result = {}

    if cache_keys is None:
        # need to itearete over a copy of the keys - list() makes a copy
        cache_keys = list(task.host.data.get("_hcache_keys_", []))

    log.debug(
        "nornir-salt:salt_clear_hcache removing hcache keys '{}'".format(cache_keys)
    )

    # iterate over given cache keys and clean them up from data
    for key in cache_keys:
        if key in task.host.data and key in task.host.data.get("_hcache_keys_", []):
            _ = task.host.data.pop(key)
            _ = task.host.data["_hcache_keys_"].remove(key)
            log.debug(
                "nornir-salt:salt_clear_hcache removed hcache key '{}'".format(key)
            )
            result[key] = True
        else:
            result[key] = False

    return Result(host=task.host, result=result)



"""
HTTPPlugin
##########

HTTP Connection plugin to interact with devices over HTTP/HTTPS.

HTTPPlugin reference
====================

.. autofunction:: nornir_salt.plugins.connections.HTTPPlugin.HTTPPlugin
"""
from typing import Any, Dict, Optional
from nornir.core.configuration import Config

CONNECTION_NAME = "http"


class HTTPPlugin:
    """
    This plugin connects to the device via HTTP using Python
    requests library.

    Connection reference name is ``http``

    Full list of inventory extras see `here <https://docs.python-requests.org/en/latest/api/>`_

    Sample Nornir inventory::

        hosts:
          ceos1:
            hostname: 10.0.1.4
            platform: arista_eos
            groups: [lab, connection_params]

          ceos2:
            hostname: 10.0.1.5
            platform: arista_eos
            groups: [lab, connection_params]

        groups:
          lab:
            username: nornir
            password: nornir
          connection_params:
            connection_options:
              http:
                port: 80
                extras:
                  transport: http
                  verify: False
                  base_url: "http://device1.lab/api/v1/"
                  headers:
                    Content-Type: "application/yang-data+json"
                    Accept: "application/yang-data+json"

    Anything under inventory extras section passed on to
    ``requests.request(method, url, **kwargs)`` call in a form of ``**kwargs``
    except for ``transport`` and ``base_url``. Inventory parameters can be
    overridden on task call.

    ``transport`` and ``base_url`` - used to form URL to send request to if no
    absolute URL provided on task call.
    """  # noqa

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]],
        configuration: Optional[Config],
    ) -> None:
        """
        HTTP does not maintain connection open, hence this method does nothing
        apart from saving inventory parameters in connection dictionary.
        """
        self.connection = {
            "extras": extras,
            "configuration": configuration,
            "hostname": hostname,
            "username": username,
            "password": password,
            "port": port,
            "platform": platform,
        }

    def close(self) -> None:
        """
        HTTP does not maintain connection open, hence this method does nothing.
        """
        self.connection = {}



"""
files
#####

Collection of task plugins to work with files saved by ``TofileProcessor``.

Files Plugin Sample usage
=========================

Code to invoke ``files`` task plugins::

    from nornir import InitNornir

    from nornir_salt.plugins.processors.ToFileProcessor import ToFileProcessor
    from nornir_salt.plugins.tasks import file_read, file_remove, file_list, file_diff, files
    from nornir_netmiko import netmiko_send_command

    nr = InitNornir(config_file="config.yaml")

    # save results to files
    nr_with_tf = nr.with_processors(
        [ToFileProcessor(tf="config_for_read", base_url="./tofile_outputs/")]
    )
    nr_with_tf.run(
        task=netmiko_send_command,
        command_string="show run"
    )

    # retrieve saved files content on demand
    res = nr.run(
        task=file_read,
        filegroup="config_for_read",
        base_url="./tofile_outputs/",
    )

Files Plugin API reference
==========================

file_read
+++++++++
.. autofunction:: nornir_salt.plugins.tasks.files.file_read

file_list
+++++++++
.. autofunction:: nornir_salt.plugins.tasks.files.file_list

file_remove
+++++++++++
.. autofunction:: nornir_salt.plugins.tasks.files.file_remove

file_diff
+++++++++
.. autofunction:: nornir_salt.plugins.tasks.files.file_diff

files dispatcher function
+++++++++++++++++++++++++
.. autofunction:: nornir_salt.plugins.tasks.files.files
"""
import os
import json
import difflib
import logging

from nornir.core.task import Result

log = logging.getLogger(__name__)


def _read_content(
    task, task_name: str, task_details: dict, data: str, timestamp: str, group: str
) -> None:
    """
    Helper function to re-use code. Gets subset of task results from
    overall data and adds it to task.results

    :param task: (obj) Nornir task object
    :param task_name: (str) Name of the task to process
    :param task_details: (dict) details of task from ToFileProcessor index file
    :param data: (str) content of file with previous results
    :param group: (str) ``tf`` files group name
    :returns: None
    """
    start, end = task_details["span"]
    # "end - 1" to account for ToFileProcessor adds "+\n" to the end
    content = data[start : end - 1]
    if task_details["content_type"] == "json":
        content = json.loads(content)
    task.results.append(
        Result(
            host=task.host,
            result=content,
            name=task_name,
            timestamp=timestamp,
            filegroup=group,
        )
    )


def _load_index_data(base_url, index):
    """
    Helper function to load json index file and return python dictionary.

    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Dictionary of index data
    """
    index_file = os.path.join(base_url, "tf_index_{}.json".format(index))
    with open(index_file, mode="r", encoding="utf-8") as f:
        index_data = json.loads(f.read())
    return index_data


def file_read(
    task,
    filegroup,
    base_url: str = "/var/nornir-salt/",
    task_name: str = None,
    last: int = 1,
    index: str = "common",
    **kwargs
):
    """
    Function to read text files content saved by ``ToFileProcessor``.

    This task reconstructs previously saved results and returns Nornir
    Result objects.

    :param filegroup: (str or list) ``tf`` group or list of ``tf`` file group names of files to load
    :param base_url: (str) OS path to folder with saved files, default "/var/nornir-salt/"
    :param last: (int) what file to read, default is 1 - the most current one
    :param task_name: (str) name of task to read previous results for, returns all results
        if ``task_name`` is empty.
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Result object with file read content
    """
    # run sanity checks
    if not filegroup:
        raise RuntimeError("nornir-salt:file_read bad filegroup '{}'".format(filegroup))

    # load index data
    index_data = _load_index_data(base_url, index)

    filegrous = [filegroup] if isinstance(filegroup, str) else filegroup

    for group in filegrous:
        # do sanity check
        if group not in index_data:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:file_read '{}' files not found".format(
                        group
                    ),
                )
            )
            continue
        if task.host.name not in index_data[group]:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    filegroup=group,
                    exception="nornir-salt:file_read '{}' host, '{}' files not found".format(
                        task.host.name, group
                    ),
                )
            )
            continue

        # get previous results metadata
        res_index = min(last - 1, len(index_data[group][task.host.name]) - 1)
        prev_res_details = index_data[group][task.host.name][res_index]
        timestamp = prev_res_details["timestamp"]

        # load file content
        with open(prev_res_details["filename"], mode="r", encoding="utf-8") as f:
            data = f.read()

        # check if need to load results for certain task only
        if task_name:
            task_details = prev_res_details["tasks"][task_name]
            _read_content(task, task_name, task_details, data, timestamp, group)
        # load content of all tasks reconstructing them in Nornir result objects
        else:
            for _task_name, task_details in prev_res_details["tasks"].items():
                _read_content(task, _task_name, task_details, data, timestamp, group)

    return Result(host=task.host, skip_results=True)


def file_list(
    task,
    filegroup=None,
    base_url: str = "/var/nornir-salt/",
    index: str = "common",
    **kwargs
):
    """
    Function to produce a list of text files saved by ``ToFileProcessor``

    :param filegroup: (str or list) ``tf`` group or list of ``tf`` file group names of files to list
    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Nornir Result object with files list
    """
    filegroup = filegroup or []
    ret = []

    # load index data
    index_data = _load_index_data(base_url, index)

    if filegroup:
        filegroups = [filegroup] if isinstance(filegroup, str) else filegroup
    else:
        filegroups = list(index_data.keys())

    for group in filegroups:
        # do sanity check
        if group not in index_data:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:file_list '{}' files group not found".format(
                        group
                    ),
                )
            )
            continue

        # iterate over previous results
        for index, file_details in enumerate(index_data[group].get(task.host.name, [])):
            tasks = file_details.pop("tasks")
            ret.append(
                {
                    "host": task.host.name,
                    "filegroup": group,
                    "last": index + 1,
                    "tasks": "\n".join(list(tasks.keys())),
                    **file_details,
                }
            )

    return Result(host=task.host, result=ret)


def file_remove(
    task,
    filegroup,
    base_url: str = "/var/nornir-salt/",
    index: str = "common",
    **kwargs
):
    """
    Function to remove files saved by ``ToFileProcessor``

    :param filegroup: (str or list or bool) ``tf`` or list of ``tf`` file group names of files
        to remove. If set to True will remove all files
    :param base_url: (str) OS path to folder with saved files, default "/var/nornir-salt/"
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Result object with removed files summary
    """
    ret = []

    # load index data
    index_data = _load_index_data(base_url, index)

    # check if need to remove files for all filegroups
    if filegroup is True:
        filegroups = list(index_data.keys())
    else:
        filegroups = filegroup if isinstance(filegroup, list) else [filegroup]

    for group in filegroups:
        # do sanity check
        if group not in index_data:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:file_remove '{}' files group not found".format(
                        group
                    ),
                )
            )
            continue

        # iterate over previous results
        for file_details in index_data[group].get(task.host.name, []):
            filename = file_details["filename"]
            tasks = file_details.pop("tasks")
            if os.path.exists(filename):
                os.remove(filename)
                ret.append(
                    {
                        "host": task.host.name,
                        "filegroup": group,
                        "tasks": "\n".join(list(tasks.keys())),
                        **file_details,
                    }
                )

        # clean up index data
        _ = index_data[group].pop(task.host.name, None)

    # save new index data
    index_file = os.path.join(base_url, "tf_index_{}.json".format(index))
    with open(index_file, mode="w", encoding="utf-8") as f:
        f.write(
            json.dumps(index_data, sort_keys=True, indent=4, separators=(",", ": "))
        )

    return Result(host=task.host, result=ret)


def file_diff(
    task,
    filegroup,
    base_url: str = "/var/nornir-salt/",
    task_name: str = None,
    last=None,
    index: str = "common",
    **kwargs
):
    """
    Function to read text files content saved by ``ToFileProcessor`` and
    return difference.

    :param filegroup: (str or list) ``tf`` file group name of files to use for diff, if list,
        runs difference for each filegroup
    :param base_url: (str) OS path to folder with saved files, default "/var/nornir-salt/"
    :param last: (int or list or str) files to diff, default is - [1, 2] - last 1 and last 2
    :param task_name: (str) name of task to read previous results for, diffs all results
        if ``task_name`` is empty.
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Result object with files difference, if files are identical reslt is True
    """
    last = last or [1, 2]

    # load index data
    index_data = _load_index_data(base_url, index)

    # get indexes of files to diff
    if isinstance(last, int):
        new = 1
        old = last
    elif isinstance(last, list) and len(last) == 2:
        new, old = last
    elif isinstance(last, str) and "," in last:
        new, old = [int(i.strip()) for i in last.split(",")][:2]
    else:
        raise TypeError(
            "nornir_data:files:diff last is not list or int or string but {}, last: {}".format(
                type(last), last
            )
        )

    filegroups = filegroup if isinstance(filegroup, list) else [filegroup]

    # iterate over filegroups and produce diffs
    for group in filegroups:
        # run sanity checks
        if group not in index_data:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:files:diff {} files not found".format(group),
                    name=group,
                )
            )
            continue
        if task.host.name not in index_data[group]:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:files:diff {} host, {} files not found".format(
                        task.host.name, group
                    ),
                    name=group,
                )
            )
            continue
        # get previous results metadata
        new_res_index = min(new - 1, len(index_data[group][task.host.name]) - 1)
        old_res_index = min(old - 1, len(index_data[group][task.host.name]) - 1)

        # check if new and old reference same file
        if new_res_index == old_res_index:
            task.results.append(
                Result(
                    host=task.host,
                    result=None,
                    exception="nornir-salt:files:diff new and old files are same - last {}, last {}".format(
                        new_res_index + 1, old_res_index + 1
                    ),
                    name=group,
                )
            )
            continue

        new_res_details = index_data[group][task.host.name][new_res_index]
        old_res_details = index_data[group][task.host.name][old_res_index]

        # load files content
        with open(new_res_details["filename"], mode="r", encoding="utf-8") as f:
            new_res_data = f.read()
        with open(old_res_details["filename"], mode="r", encoding="utf-8") as f:
            old_res_data = f.read()

        # if task_name given, retrieve task results content
        if task_name:
            # get task details
            new_task_details = new_res_details["tasks"][task_name]
            old_task_details = old_res_details["tasks"][task_name]
            # get task span
            new_start, new_end = new_task_details["span"]
            old_start, old_end = old_task_details["span"]
            # get task text results to diff
            new_todiff_data = new_res_data[new_start : new_end - 1]
            old_todiff_data = old_res_data[old_start : old_end - 1]
        # use all data for diff
        else:
            new_todiff_data = new_res_data
            old_todiff_data = old_res_data

        # run diff
        difference = difflib.unified_diff(
            a=old_todiff_data.splitlines(keepends=True),
            b=new_todiff_data.splitlines(keepends=True),
            fromfile=old_res_details["filename"],
            tofile=new_res_details["filename"],
        )

        res = "".join(difference)
        res = res if res else True
        task.results.append(Result(host=task.host, result=res, name=group))

    return Result(host=task.host, skip_results=True)


def files(task, call, *args, **kwargs):
    """
    Dispatcher function to call one of the functions.

    :param call: (str) nickname of function to call
    :param arg: (list) function arguments
    :param kwargs: (dict) function key-word arguments
    :return: function execution results

    Call function nicknames:

    * ``ls`` - calls `file_list`_
    * ``rm`` - calls `file_remove`_
    * ``read`` - calls `file_read`_
    * ``diff`` - calls `file_diff`_
    """
    dispatcher = {
        "ls": file_list,
        "rm": file_remove,
        "read": file_read,
        "diff": file_diff,
    }
    return dispatcher[call](task, *args, **kwargs)



"""
ResultSerializer
################

Helper function to transform Nornir results object in python dictionary to
ease programmatic consumption or further transformation in other formats
such as JSON or YAML

ResultSerializer supports serialization of results of these object types::

    list, tuple, dict, str, int, bool, set, type(None)

If task result is not one of above types, it is converted to string.

Exception object transformed to string.

ResultSerializer Sample Usage
=============================

Code to demonstrate how to invoke ResultSerializer::

    from nornir import InitNornir
    from nornir_netmiko import netmiko_send_command
    from nornir_salt.plugins.functions import ResultSerializer

    nr = InitNornir(config_file="config.yaml")

    result = NornirObj.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

    result_dictionary = ResultSerializer(result, add_details=True)

    # work further with result_dictionary
    # ...

ResultSerializer returns
========================

ResultSerializer capable of returning two different structures, each one
can contain additional task details. The difference between structures is
in the way how tasks are represented.

First structure uses dictionary keyed by task name, where values are
task's results.

Second structure type uses list to store task results.

If ``add_details`` is False and ``to_dict`` is True returns dictionary::

    {
        "hostname_1": {
            "task_name_1": result,
            "task_name_2": result
        },
        "hostname_2": {
            "task_name_1": result,
            "task_name_2": result
        }
    }

For instance::

    {'IOL1': {'show clock': '*00:55:21.236 EET Tue Feb 9 2021',
               'show run | inc hostname': 'hostname IOL1'},
     'IOL2': {'show clock': '*00:55:21.234 EET Tue Feb 9 2021',
               'show run | inc hostname': 'hostname IOL2'}}

If ``add_details`` is True and ``to_dict`` is True returns dictionary
with additional details::

    {
        "hostname_1": {
            "task_name_1": {
                "changed": False,
                "diff: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            },
            "task_name_2": {
                "changed": False,
                "diff: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            }
        },
        "hostname_2": {
            "task_name_1": {
                "changed": False,
                "diff}: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            }
        }
    }

For example::

    {'IOL1': {'show clock': {'changed': False,
                             'diff': '',
                             'exception': 'None',
                             'failed': False,
                             'result': '*00:57:45.398 EET Tue Feb 9 2021'},
              'show run | inc hostname': {'changed': False,
                                          'diff': '',
                                          'exception': 'None',
                                          'failed': False,
                                          'result': 'hostname IOL1'}},
     'IOL2': {'show clock': {'changed': False,
                             'diff': '',
                             'exception': 'None',
                             'failed': False,
                             'result': '*00:57:45.489 EET Tue Feb 9 2021'},
              'show run | inc hostname': {'changed': False,
                                          'diff': '',
                                          'exception': 'None',
                                          'failed': False,
                                          'result': 'hostname IOL2'}}}

If ``add_details`` is False and ``to_dict`` is False returns dictionary::

    {
        "hostname_1": [
            {"name": "task_name_1", "result": result},
            {"name": "task_name_2", "result": result}
        ],
        "hostname_2": [
            {"name": "task_name_1", "result": result},
            {"name": "task_name_2", "result": result}
        ]
    }

If ``add_details`` is True and ``to_dict`` is False returns dictionary::

    {
        "hostname_1": [
            {
                "name": "task_name_1",
                "changed": False,
                "diff: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            },
            {
                "name": "task_name_2",
                "changed": False,
                "diff: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            }
        ],
        "hostname_2": [
            {
                "name": "task_name_1",
                "changed": False,
                "diff: "",
                "exception": None,
                "failed": False,
                "result": "result string"
            }
        ]
    }

Skipping results
================

ResultSerializer by default skips all tasks with name starting with
underscore ``_``, in addition results skipped if ``Result`` object
contains ``skip_results`` attribute and it set to ``True``.

Above skip logic ignored if ``Result`` object exception is not empty.

ResultSerializer reference
==========================

.. autofunction:: nornir_salt.plugins.functions.ResultSerializer.ResultSerializer
"""
import logging
from nornir.core.task import AggregatedResult

log = logging.getLogger(__name__)

supported_types = [list, tuple, dict, str, int, bool, set, type(None)]


def ResultSerializer(nr_results, add_details=False, to_dict=True, skip=None):
    """
    :param nr_results: ``nornir.core.task.AggregatedResult`` object
    :param add_details: boolean to indicate if results should contain more info, default
        is False
    :param to_dict: (bool) default is True, forms nested dictionary structure, if False
        forms results in a list.
    :param skip: (list) list of Result object attributes names to omit, default is
        "severity_level", "stderr", "stdout", "host"
    """
    skip = skip or ["severity_level", "stderr", "stdout", "host"]
    # run check
    if not isinstance(nr_results, AggregatedResult):
        return nr_results

    # form nested dictionary structure
    if to_dict:
        ret = {}
        for hostname, results in nr_results.items():
            for i in results:
                exception = str(i.exception) if i.exception is not None else None
                # skip tasks such as _task_foo_bar unless exception
                if i.name and i.name.startswith("_") and not exception:
                    continue
                # skip tasks if signaled to do so
                elif (
                    hasattr(i, "skip_results")
                    and i.skip_results is True
                    and not exception
                ):
                    continue
                # add hostname to results
                ret.setdefault(hostname, {})
                # add results details if requested to do so
                if add_details:
                    ret[hostname][i.name] = {
                        k: v
                        for k, v in vars(i).items()
                        if k not in skip and type(v) in supported_types
                    }
                    ret[hostname][i.name].setdefault("result", str(i.result))
                    ret[hostname][i.name]["failed"] = True if exception else i.failed
                    ret[hostname][i.name]["exception"] = exception
                    ret[hostname][i.name].pop("name")
                # form results for the rest of tasks
                else:
                    ret[hostname][i.name] = (
                        i.result if type(i.result) in supported_types else str(i.result)
                    )

    # form plain list of results
    else:
        ret = []
        for hostname, results in nr_results.items():
            for i in results:
                exception = str(i.exception) if i.exception is not None else None
                # skip group tasks such as _task_foo_bar unless exception
                if i.name and i.name.startswith("_") and not exception:
                    continue
                # skip tasks if signalled to do so
                elif (
                    hasattr(i, "skip_results")
                    and i.skip_results is True
                    and not exception
                ):
                    continue
                # add results details if requested to do so
                elif add_details:
                    ret.append(
                        {
                            k: v
                            for k, v in vars(i).items()
                            if k not in skip and type(v) in supported_types
                        }
                    )
                    ret[-1].setdefault("result", str(i.result))
                    ret[-1]["failed"] = True if exception else i.failed
                    ret[-1]["exception"] = exception
                    ret[-1]["host"] = i.host.name
                # form results for the rest of tasks
                else:
                    ret.append(
                        {
                            "host": hostname,
                            "name": i.name,
                            "result": i.result
                            if type(i.result) in supported_types
                            else str(i.result),
                        }
                    )

    return ret



"""
salt_cfg_gen
############

This is a function to test configuration generation without applying
it to devices. It goes through full configuration rendering and helps
to verify configuration content or simply generate it.

Used in conjunction with ``nr.cfg_gen`` execution module function.

salt_cfg_gen reference
======================

.. autofunction:: nornir_salt.plugins.tasks.salt_cfg_gen.salt_cfg_gen
"""
from nornir.core.task import Result


def salt_cfg_gen(task, config=None, **kwargs):
    """
    Task function for ``nr.cfg_gen`` function to return template rendered
    with pillar and Nornir host's Inventory data.

    In essence, this function echoes back content of ``config`` argument.

    :param config: (str) configuration string to return
    :return result: Nornir result object with task execution results
    """
    task.name = "salt_cfg_gen"

    # get configuration
    if "config" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["config"]
    elif "commands" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["filename"]

    # transform configuration to string if list/tuple given
    if isinstance(config, (list, tuple)):
        config = "\n".join(config)

    return Result(host=task.host, result=config)



"""
N2G Processor Plugin
####################

Processor plugin that uses N2G library to produce diagrams out of
network devices output.

N2GProcessor Sample Usage
=========================

Code to demonstrate how to use ``N2GProcessor`` plugin::

    TBD

N2GProcessor reference
=========================

.. autofunction:: nornir_salt.plugins.processors.N2GProcessor.N2GProcessor
"""
import logging

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task


log = logging.getLogger(__name__)


class N2GProcessor:
    """
    N2GProcessor can generate diagrams out network devices output.

    :param module: (str) name of diagramming module to use
    """

    def __init__(self):
        pass

    def task_started(self, task: Task) -> None:
        pass  # ignore

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore subtasks

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass  # ignore subtasks

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass



"""
PyATSUnicon
###########

`PyATS <https://pubhub.devnetcloud.com/media/pyats/docs/index.html>`_
connection plugin to interact with network devices.

Check `supported platofrms <https://pubhub.devnetcloud.com/media/unicon/docs/user_guide/supported_platforms.html>`_
page for platform codes.

This plugin uses Genie to load testbed inventory with single device and optional jumphosts definition,
Genie itslef relies on PyATS to perform lower level tasks, where PyATS relies on Unicon library to
communicate with devices over CLI.

PyATSUnicon reference
=====================

.. autofunction:: nornir_salt.plugins.connections.PyATSUnicon.PyATSUnicon
"""
import logging
from typing import Any, Dict, Optional
from nornir.core.configuration import Config

try:
    from genie import testbed

    HAS_GENIE = True
except ImportError:
    HAS_GENIE = False

CONNECTION_NAME = "pyats"

log = logging.getLogger(__name__)


class PyATSUnicon:
    """
    This plugin makes use of PyATS tesbed definition to initiate device connections, testbed can
    be partially reconstructed out of Nornir inventory or complete testbed data can be provided under
    extras section.

    Sample minimum inventory that reconstructed to PyATS testbed::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          port: 22
          connection_options:
            pyats:
              platform: eos
              extras:
                devices:
                  host-1: {}

    ``connection_options:pyats:extras`` section used to load PyATS testbed object.

    Above invetory reconstructed to this PyATS testbed data::

        devices:
          host-1:
            os: eos
            credentials:
              default:
                username: admin
                password: admin
            connections:
              default:
                protocol: ssh
                ip: 192.168.16.20
                port: 22
              vty_1:
                protocol: ssh
                ip: 192.168.16.20

    Alternatively, full testbed data can be specified using extras like this::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          connection_options:
            pyats:
              extras:
                testbed:
                  name: eos_testbed
                devices:
                  host-1:
                    os: eos
                    credentials:
                      default:
                        username: admin
                        password: admin
                    connections:
                      default:
                        protocol: ssh
                        ip: 192.168.16.20
                        port: 22

    In that case, because all mandatory parameters ``os``, ``connections`` and ``credentials`` provided,
    extras data used as is to load PyATS tesbed.

    It is mandatory to specify exact device cli prompt under ``connection_options:pyats:extras:devices``
    as a dictionary key as well as for top level device key. In other words this will not work::

        host-1-foo:
          hostname: 192.168.16.20
          connection_options:
            pyats:
              extras:
                devices:
                  host-1-bar: {}

    Where ``host-1-foo`` is Nornir invetory host name and ``host-1-bar`` is actual device prompt
    as seen on cli, these two keys must be of the same value.

    This plugin establishes all connections to device on startup.

    To use connections pool instead of single connection, need to provide ``pool`` argument integer
    of value more or equal to 2 in connection's parameters, otherwise connection ignored::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          connection_options:
            pyats:
              extras:
                devices:
                  host-1:
                    os: eos
                    connections:
                      default:
                        protocol: ssh
                        ip: 192.168.16.20
                        pool: 3

    ``pool`` connection argument is not part of PyATS native testbed schema and only
    relevant in PyATSUnicon plugin context.
    """  # noqa

    def open(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
        platform: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        extras = extras or {}
        pools = {}  # keyed by device name, connections name and pool_size

        # check if need to add missing data to devices in testbed
        for device_name, device_data in extras["devices"].items():
            if "credentials" not in device_data:
                device_data["credentials"] = {
                    "default": {"username": username, "password": password}
                }
            if "connections" not in device_data:
                device_data["connections"] = {
                    "default": {"protocol": "ssh", "ip": hostname, "port": port or 22}
                }
            if "telnet" in platform:
                device_data["connections"]["default"]["protocol"] = "telnet"
            if "os" not in device_data and platform:
                device_data["os"] = platform

            # add/extract data to/from connections
            for connection_name, connection_data in device_data["connections"].items():
                # check if connection is a pool
                pool_size = int(connection_data.pop("pool", 0))
                if pool_size > 1:
                    pools.setdefault(device_name, {})
                    pools[device_name][connection_name] = pool_size
                # check if connection missing "ip" field
                if "ip" not in connection_data and hostname:
                    connection_data["ip"] = hostname

        # load testbed
        self.connection = testbed.load(extras)

        # initiate non pool connections
        for device_name, device_data in extras["devices"].items():
            for connection_name, connection_data in device_data["connections"].items():
                # check if need to skip as there is connections pool exists already
                if not pools.get(device_name, {}).get(connection_name):
                    self.connection.devices[device_name].connect(
                        alias=connection_name, via=connection_name
                    )

        # initiate pool connections if any
        for device_name, device_data in pools.items():
            for connection_name, pool_size in device_data.items():
                self.connection.devices[device_name].start_pool(
                    size=pool_size, via=connection_name, alias=connection_name
                )
                # set pool size attribute to connections pool object for tasks to make use of it
                conn = getattr(self.connection.devices[device_name], connection_name)
                conn._pool_size = pool_size

    def close(self) -> None:
        # destroy connections to all devices in this testbed
        self.connection.destroy()



"""
pyats_send_commands
#########################

This task plugin uses PyATS devices ``execute`` method
to send multiple commands to devices pre-processing commands accordingly.

Pre-processing includes:

- Check and if any - retrieve per-host commands from host's inventory data
  ``task.host.data["__task__"]["commands"]`` or from ``task.host.data["__task__"]["filename"]``
- If command is a multi-line string, split it to individual lines or form a list with single command
- Iterate over commands list and remove empty strings
- Iterate over commands and replace ``new_line_char`` with ``\\n`` new line

Tere are several modes that ``pyats_send_commands`` task plugin can operate in:

1. If ``parse`` is true and device platform supports it, send commands one by one parsing
   their output waiting for ``interval`` in betwen commands if ``interval`` provided, if
   no parser available for this platofrm to parse command output, exception message returned
   for such a command
2. If ``interval`` argument provided, commands send one by one to device using
   ``execute`` method sleeping for given ``interval`` between commands
3. If ``via`` argument refers to connections pool object, send commands in parallel,
   commands excution order not guaranteed
4. By default, all commands supplied to device's connection ``execute`` method as is

Dependencies:

* `PyATS library <https://pypi.org/project/pyats/>`_ required
* `Genie library <https://pypi.org/project/genie/>`_ required

Useful links:

- `Genie parsers <https://pubhub.devnetcloud.com/media/genie-feature-browser/docs/#/parsers>`_

Sample Usage
============

Given this inventory::

    host-1:
      connection_options:
        pyats:
          extras:
            devices:
              host-1:
                os: iosxe
                credentials:
                  default:
                    username: nornir
                    password: nornir
                connections:
                  default:
                    protocol: ssh
                    ip: 10.0.1.4
                    port: 22
                  vty_1:
                    protocol: ssh
                    ip: 10.0.1.4
                    pool: 3

Code to invoke ``pyats_send_commands`` task::

    from nornir_salt import pyats_send_commands

    # send via "default" connection
    output_via_default = nr.run(
        task=pyats_send_commands,
        commands=["show run", "show clock"]
    )

    # send via vty_1 connection pool of 3 SSH connections
    output_via_pool = nr.run(
        task=pyats_send_commands,
        commands=["show run", "show clock"],
        via="vty_1"
    )

    # send via "default" connection with 5s interval between commands
    output_with_interval = nr.run(
        task=pyats_send_commands,
        commands=["show run", "show clock"],
        interval=5
    )

    # send commands and parse output
    output_parse = nr.run(
        task=pyats_send_commands,
        commands=["show hostname", "show clock"],
        parse=True
    )

Returns Nornir results object with individual tasks names set
equal to commands sent to device.

Reference
=========

.. autofunction:: nornir_salt.plugins.tasks.pyats_send_commands.pyats_send_commands
"""
import time
import logging
import traceback

from nornir.core.task import Result, Task
from concurrent.futures import ThreadPoolExecutor, as_completed

try:
    from pyats.connections.pool import ConnectionPool
    from genie.libs.parser.utils.common import ParserNotFound
    from genie.libs.parser.utils import get_parser

    HAS_PYATS = True
except ImportError:
    HAS_PYATS = False

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "pyats"


def _form_results(task, res, command):
    """
    Helper function to save code on forming results out of PyATS output

    :param task: (obj) Nornir task object
    :param res: (str, dict) PyATS command execution results
    :param command: (str, list) command string or list of commands sent
    """
    # if non-multiline or single command was provided res is a string
    if isinstance(res, str):
        cmd = ", ".join(command) if isinstance(command, list) else command.strip()
        task.results.append(Result(host=task.host, result=res, name=cmd))
    # if multiline command provided or a list of commands res is a dictionary
    elif isinstance(res, dict):
        for cmd, output in res.items():
            task.results.append(Result(host=task.host, result=output, name=cmd.strip()))


def pyats_send_commands(
    task: Task,
    commands: list = None,
    interval: int = None,
    new_line_char: str = "_br_",
    via: str = "default",
    parse: bool = False,
    **kwargs
):
    """
    Nornir Task function to send show commands to devices using PyATS Unicon module.

    Per-host ``commands`` can be provided using host's object ``data`` attribute
    with ``__task__`` key with value set to dictionary with ``commands`` key
    containing a list of or a multiline string of commands to send to device, e.g.::

        print(host.data["__task__"]["commands"])

        ["ping 1.1.1.1 source 1.1.1.2", "show clock"]

    Alternatively, ``__task__`` can contain ``filename`` key with commands string
    to send to device.

    :param kwargs: (dict) used with connection's ``execute`` or testbed's ``parse`` method as ``**kwargs``
    :param commands: (list or str) list or multiline string of commands to send to device
    :param interval: (int) interval between sending commands, default is None - commands send simelteniously
    :param new_line_char: (str) characters to replace in commands with new line ``\\n``
        before sending command to device, default is ``_br_``, useful to simulate enter key
    :param via: (str) testbed inventory connection name, default is ``default``
    :param parse: (bool) if True, parses command output and returns structured data
    :return result: Nornir result object with task results named after commands
    """
    # run sanity check
    if not HAS_PYATS:
        return Result(
            host=task.host,
            failed=True,
            exception="Failed to import PyATS library, is it installed?",
        )

    commands = commands or []

    # get per-host commands if any
    if "commands" in task.host.data.get("__task__", {}):
        if commands:
            for c in task.host.data["__task__"]["commands"]:
                if c not in commands:
                    commands.append(c)
        else:
            commands = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        commands = task.host.data["__task__"]["filename"]

    # normalize commands to a list
    if isinstance(commands, str):
        commands = commands.splitlines()
    elif isinstance(commands, str):
        commands = [commands]

    # remove empty lines/commands that can left after rendering
    commands = [c for c in commands if c.strip()]

    # iterate over commands and see if need to add empty line - hit enter
    commands = [
        c.replace(new_line_char, "\n") if new_line_char in c else c for c in commands
    ]

    # get PyATS testbed, device and connection objects
    testbed = task.host.get_connection(CONNECTION_NAME, task.nornir.config)
    device = testbed.devices[task.host.name]
    if hasattr(device, via):
        connection = getattr(device, via)
    else:
        raise RuntimeError("{} has no connection '{}'".format(task.host.name, via))

    # check if nee to parse output
    if parse:
        log.debug(
            "nornir-salt:pyats_send_commands sending commands one by one and parsing output"
        )
        for command in commands:
            try:
                get_parser(command, device)  # raises exception if no parse available
                output = connection.execute(command, **kwargs)
                result = device.parse(command, output=output)
                result = result.q.reconstruct()  # result is PyATS Dq Dict object
                exception = None
            except ParserNotFound as e:
                result = str(e)
                exception = traceback.format_exc()
            task.results.append(
                Result(
                    host=task.host,
                    result=result,
                    exception=exception,
                    name=command.strip(),
                )
            )
            if isinstance(interval, (int, float)):
                time.sleep(interval)
    # send commands one by one with interval
    elif isinstance(interval, (int, float)):
        log.debug(
            "nornir-salt:pyats_send_commands connection '{}', sending commands with '{}s' interval".format(
                via, interval
            )
        )
        for command in commands:
            _form_results(task, connection.execute(command, **kwargs), command)
            time.sleep(interval)
    # make use of PyATS connections pool to send commands in parrallel
    elif isinstance(connection, ConnectionPool):
        size = connection._pool_size if hasattr(connection, "_pool_size") else 5
        log.debug(
            "nornir-salt:pyats_send_commands connections pool '{}', size '{}', sending commands in parrallel".format(
                via, size
            )
        )
        # execute threads to send commands across multiple connections
        with ThreadPoolExecutor(size) as pool:
            cmd_futures = {
                pool.submit(lambda cmd: connection.execute(cmd, **kwargs), cmd): cmd
                for cmd in commands
            }
            for future in as_completed(cmd_futures):
                _form_results(task, future.result(), cmd_futures[future])
    # send all commands at once
    else:
        log.debug(
            "nornir-salt:pyats_send_commands connection '{}', sending commands all at once".format(
                via
            )
        )
        _form_results(task, connection.execute(commands, **kwargs), commands)

    # set skip_results to True, for ResultSerializer to ignore
    # results for grouped task itself, which are usually None
    return Result(host=task.host, skip_results=True)
